# -*- coding: utf-8 -*-
import datetime
import warnings
import pandas as pd
import numpy as np

from rqdatac.services.basic import instruments
from rqdatac.services.calendar import get_next_trading_date, is_trading_date, get_previous_trading_date
from rqdatac.services.live import get_ticks

from rqdatac.validators import (
    ensure_string,
    ensure_list_of_string,
    check_items_in_container,
    ensure_instruments,
    ensure_date_range,
    is_panel_removed,
    raise_for_no_panel,
)
from rqdatac.utils import (
    to_date_int,
    to_datetime,
    to_date,
    int8_to_datetime,
    int17_to_datetime_v,
    int17_to_datetime,
    today_int,
    date_to_int8,
    string_types
)
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api, ttl_cache, compatible_with_parm
from rqdatac.share.errors import PermissionDenied, MarketNotSupportError


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_price(
        order_book_ids,
        start_date=None,
        end_date=None,
        frequency="1d",
        fields=None,
        adjust_type="pre",
        skip_suspended=False,
        expect_df=True,
        market="cn",
        **kwargs
):
    """获取证券的历史数据

    :param order_book_ids: 股票列表
    :param market: 地区代码, 如 'cn' (Default value = "cn")
    :param start_date: 开始日期, 如 '2013-01-04' (Default value = None)
    :param end_date: 结束日期, 如 '2014-01-04' (Default value = None)
    :param frequency: 可选参数, 默认为日线。日线使用 '1d', 分钟线 '1m' (Default value = "1d")
    :param fields: 可选参数。默认为所有字段。 (Default value = None)
    :param adjust_type: 可选参数,默认为‘pre', 返回开盘价，收盘价，最高价，最低价依据get_ex_factor 复权因子（包含分红，拆分），volume依据get_split 复权因子（仅涵盖拆分）计算的前复权数据
            'none'将返回原始数据
            'post'返回开盘价，收盘价，最高价，最低价依据get_ex_factor 复权因子（包含分红，拆分），volume依据get_split 复权因子（仅涵盖拆分）计算的后复权数据
            'pre_volume'返回开盘价，收盘价，最高价，最低价,成交量依据get_ex_factor 复权因子（包含分红，拆分）计算的前复权数据
            'post_volume'返回开盘价，收盘价，最高价，最低价,成交量依据get_ex_factor 复权因子（包含分红，拆分）计算的后复权数据
            'internal'返回只包含拆分的前复权数据。 (Default value = "pre")
    :param skip_suspended: 可选参数，默认为False；当设置为True时，返回的数据会过滤掉停牌期间，
                    此时order_book_ids只能设置为一只股票 (Default value = False)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :returns: 如果仅传入一只股票, 返回一个 pandas.DataFrame
        如果传入多只股票, 则返回一个 pandas.Panel

    """
    # tick数据
    if frequency == "tick":
        return get_tick_price(order_book_ids, start_date, end_date, fields, expect_df, market)
    elif frequency.endswith(("d", "m", "w")):
        duration = int(frequency[:-1])
        frequency = frequency[-1]
        assert 1 <= duration <= 240, "frequency should in range [1, 240]"
        if market == "hk" and frequency == "m" and duration not in (1, 5, 15, 30, 60):
            raise ValueError("frequency should be str like 1m, 5m, 15m 30m,or 60m")
        elif frequency == 'w' and duration not in (1,):
            raise ValueError("Weekly frequency should be str '1w'")
    else:
        raise ValueError("frequency should be str like 1d, 1m, 5m or tick")
    # 验证adjust_type
    if "adjusted" in kwargs:
        adjusted = kwargs.pop("adjusted")
        adjust_type = "pre" if adjusted else "none"

    if kwargs:
        raise ValueError('unknown kwargs: {}'.format(kwargs))

    valid_adjust = ["pre", "post", "none", "pre_volume", "post_volume"]
    ensure_string(adjust_type, "adjust_type")
    check_items_in_container(adjust_type, valid_adjust, "adjust_type")
    order_book_ids = ensure_list_of_string(order_book_ids, "order_book_ids")
    if skip_suspended and len(order_book_ids) > 1:
        raise ValueError("only accept one order_book_id or symbol if skip_suspended is True")

    assert isinstance(skip_suspended, bool), "'skip_suspended' should be a bool"
    assert isinstance(expect_df, bool), "'expect_df' should be a bool"

    order_book_ids, stocks, funds, indexes, futures, futures888, spots, options, convertibles, repos = classify_order_book_ids(
        order_book_ids, market
    )
    if not order_book_ids:
        warnings.warn("no valid instrument")
        return
    start_date, end_date = _ensure_date(
        start_date, end_date, stocks, funds, indexes, futures, spots, options, convertibles, repos
    )
    from rqdatac.services.detail.get_price_df import get_price_df, get_week_df
    if frequency != 'w':
        df = get_price_df(
            order_book_ids, start_date, end_date, frequency, duration, fields, adjust_type, skip_suspended,
            stocks, funds, indexes, futures, futures888, spots, options, convertibles, repos, market
        )
        if df is None or expect_df:
            return df
        # 单个合约
        if len(df.index.get_level_values(0).unique()) == 1:
            df.reset_index(level=0, inplace=True, drop=True)
            # df.index.name = None
            if len(df.columns) == 1:
                df = df[df.columns[0]]
            return df
        # 单个字段
        elif len(df.columns) == 1:
            field = df.columns[0]
            df = df.unstack(0)[field]
            # df.index.name = None
            df.columns.name = None
            return df
        raise_for_no_panel(False)
        warnings.warn("Panel is removed after pandas version 0.25.0."
                      " the default value of 'expect_df' will change to True in the future.")
        # 交换index的顺序，以制作panel
        return df.swaplevel().to_panel()
    else:
        if not expect_df:
            raise ValueError(
                "Weekly frequency can only return a DataFrame object, set 'expect_df' to True to resolve this")
        if skip_suspended:
            raise ValueError(
                "Weekly frequency does not support skipping suspended trading days, set 'skip_suspended' to False to resolve this")
        start_date, end_date = _weekly_start_end_date_handler(start_date, end_date)
        if start_date > end_date:
            # 如果*当周没有结束*
            # 或者start date 和 end date 不能涵盖当周所有的交易日，查询该周的数据时返回为空。
            return None
        return get_week_df(order_book_ids, start_date, end_date, fields, adjust_type, market,
                           *(stocks, funds, indexes, futures, futures888, spots, options, convertibles, repos))


def _ensure_date(start_date, end_date, stocks, funds, indexes, futures, spots, options, convertibles, repos):
    default_start_date, default_end_date = ensure_date_range(start_date, end_date)

    only_futures = futures and (not stocks) and (not funds) and (not indexes) and (not spots) and (
        not options) and (not convertibles) and (not repos)
    if only_futures and len(futures) == 1:
        # 如果只有一只期货, 则给 start_date 和 end_date 合适的默认值
        # 连续合约的listed_date和de_listed_date都为0, 因此需要特殊处理
        if futures[0].listed_date != "0000-00-00":
            default_start_date = to_date_int(futures[0].listed_date)
        if futures[0].de_listed_date != "0000-00-00":
            default_end_date = to_date_int(futures[0].de_listed_date)

    start_date = to_date_int(start_date) if start_date else default_start_date
    end_date = to_date_int(end_date) if end_date else default_end_date
    if start_date < 20000104:
        warnings.warn("start_date is earlier than 2000-01-04, adjusted to 2000-01-04")
        start_date = 20000104
    return start_date, end_date


def _ensure_fields(fields, fields_dict, stocks, funds, futures, futures888, spots, options, convertibles, indexes,
                   repos):
    has_dominant_id = False
    future_only = futures and not any([stocks, funds, spots, options, convertibles, indexes, repos])
    all_fields = set(fields_dict["common"])
    if futures:
        all_fields.update(fields_dict["future"])
    if stocks:
        all_fields.update(fields_dict["stock"])
    if funds:
        all_fields.update(fields_dict["fund"])
    if spots:
        all_fields.update(fields_dict["spot"])
    if options:
        all_fields.update(fields_dict["option"])
    if convertibles:
        all_fields.update(fields_dict["convertible"])
    if indexes:
        all_fields.update(fields_dict["index"])
    if repos:
        all_fields.update(fields_dict["repo"])
    if future_only and futures888 and len(futures) == len(futures888) and not fields:
        has_dominant_id = True

    if fields:
        fields = ensure_list_of_string(fields, "fields")
        fields_set = set(fields)
        if len(fields_set) < len(fields):
            warnings.warn("duplicated fields: %s" % [f for f in fields if fields.count(f) > 1])
            fields = list(fields_set)
        # 只有期货类型
        if 'dominant_id' in fields:
            fields.remove("dominant_id")
            if not fields:
                raise ValueError("can't get dominant_id separately, please use futures.get_dominant")
            if futures888:
                has_dominant_id = True
            else:
                warnings.warn(
                    "only if one of the order_book_id is future and contains 88/888/99/889 can the dominant_id be selected in fields")
        check_items_in_container(fields, all_fields, "fields")
        return fields, has_dominant_id
    else:
        return list(all_fields), has_dominant_id


def classify_order_book_ids(order_book_ids, market):
    ins_list = ensure_instruments(order_book_ids, market=market)
    _order_book_ids = []
    stocks = []
    funds = []
    indexes = []
    futures = []
    futures_888 = {}
    spots = []
    options = []
    convertibles = []
    repos = []
    for ins in ins_list:
        if ins.order_book_id not in _order_book_ids:
            _order_book_ids.append(ins.order_book_id)
            if ins.type == "CS":
                stocks.append(ins.order_book_id)
            elif ins.type == "INDX":
                indexes.append(ins.order_book_id)
            elif ins.type in {"ETF", "LOF", "SF"}:
                funds.append(ins.order_book_id)
            elif ins.type == "Future":
                if ins.order_book_id.endswith(("888", "889")):
                    futures_888[ins.order_book_id] = ins.underlying_symbol
                futures.append(ins)
            elif ins.type == "Spot":
                spots.append(ins.order_book_id)
            elif ins.type == "Option":
                options.append(ins.order_book_id)
            elif ins.type == "Convertible":
                convertibles.append(ins.order_book_id)
            elif ins.type == "Repo":
                repos.append(ins.order_book_id)
    return _order_book_ids, stocks, funds, indexes, futures, futures_888, spots, options, convertibles, repos


def _weekly_start_end_date_handler(start_date, end_date):
    start_date = to_date(start_date)
    monday = start_date - datetime.timedelta(days=start_date.weekday())
    first_trading_day_in_week = monday if is_trading_date(monday) else get_next_trading_date(monday)
    if first_trading_day_in_week < start_date:
        start_date = monday + datetime.timedelta(weeks=1)

    end_date = to_date(end_date)
    if end_date > datetime.date.today():
        end_date = datetime.date.today()
    friday = end_date - datetime.timedelta(days=end_date.weekday()) + datetime.timedelta(days=4)
    last_trading_day_in_week = friday if is_trading_date(friday) else get_previous_trading_date(friday)
    if last_trading_day_in_week > end_date:
        end_date = friday - datetime.timedelta(weeks=1)

    return to_date_int(start_date), to_date_int(end_date)


@ttl_cache(15 * 60)
def daybar_for_tick_price(order_book_id):
    ins = instruments(order_book_id)
    today = to_date_int(datetime.datetime.today())

    if ins.type in ("Future", "Spot"):
        fields = ["prev_settlement", "open", "close", "limit_up", "limit_down"]
    elif ins.type == "Option":
        fields = ["prev_settlement", "open", "close"]
    elif ins.type in ("LOF", "ETF"):
        fields = ["open", "close", "limit_up", "limit_down", "iopv"]
    elif ins.type == "CS":
        fields = ["open", "close", "limit_up", "limit_down"]
    else:
        fields = ["open", "close"]

    df = get_price(
        ins.order_book_id,
        "2004-12-31",
        today,
        frequency="1d",
        fields=fields,
        adjust_type="none",
        skip_suspended=False,
        expect_df=is_panel_removed,
        market="cn",
    )

    if df is not None and isinstance(df.index, pd.MultiIndex):
        df.reset_index(level=0, inplace=True)
    return df


EQUITIES_TICK_FIELDS = [
    "trading_date", "open", "last", "high", "low",
    "prev_close", "volume", "total_turnover", "limit_up", "limit_down",
    "a1", "a2", "a3", "a4", "a5", "b1", "b2", "b3", "b4", "b5", "a1_v", "a2_v", "a3_v",
    "a4_v", "a5_v", "b1_v", "b2_v", "b3_v", "b4_v", "b5_v", "change_rate",
    "num_trades",
]
FUND_TICK_FIELDS = EQUITIES_TICK_FIELDS + ["iopv", "prev_iopv"]
FUTURE_TICK_FIELDS = EQUITIES_TICK_FIELDS + ["open_interest", "prev_settlement"]
EQUITIES_TICK_COLUMNS = EQUITIES_TICK_FIELDS
FUTURE_TICK_COLUMNS = [
    "trading_date", "open", "last", "high", "low", "prev_settlement",
    "prev_close", "volume", "open_interest", "total_turnover", "limit_up", "limit_down",
    "a1", "a2", "a3", "a4", "a5", "b1", "b2", "b3", "b4", "b5", "a1_v", "a2_v", "a3_v",
    "a4_v", "a5_v", "b1_v", "b2_v", "b3_v", "b4_v", "b5_v", "change_rate",
]
FUND_TICK_COLUMNS = FUND_TICK_FIELDS
RELATED_DABAR_FIELDS = {"open", "prev_settlement", "prev_close", "limit_up", "limit_down", "change_rate"}


def get_tick_price(order_book_ids, start_date, end_date, fields, expect_df, market):
    df = get_tick_price_multi_df(order_book_ids, start_date, end_date, fields, market)
    if df is not None and not expect_df and isinstance(order_book_ids, string_types):
        df.reset_index(level=0, drop=True, inplace=True)
    return df


def convert_history_tick_to_multi_df(data, dt_name, fields, convert_dt):
    line_no = 0
    dt_set = set()
    obid_level = []
    obid_slice_map = {}
    for i, (obid, d) in enumerate(data):
        dates = d.pop("date")
        if len(dates) == 0:
            continue
        times = d.pop("time")
        dts = d[dt_name] = [_convert_int_to_datetime(dt, tm) for dt, tm in zip(dates, times)]

        dts_len = len(dts)

        if not obid_level or obid_level[-1] != obid:
            obid_level.append(obid)
        obid_slice_map[(i, obid)] = slice(line_no, line_no + dts_len, None)

        dt_set.update(dts)
        line_no += dts_len

    if line_no == 0:
        return

    daybars = {}
    if set(fields) & RELATED_DABAR_FIELDS:
        ins_list = ensure_instruments(obid_level)
        for ins in ins_list:
            daybar = daybar_for_tick_price(ins.order_book_id)
            if daybar is not None:
                daybar['prev_close'] = daybar['close'].shift(1)
                if ins.type in ["ETF", "LOF"]:
                    daybar['prev_iopv'] = daybar['iopv'].shift(1)
            daybars[ins.order_book_id] = daybar
        fields_ = list(set(fields) | {"last", "volume"})
    else:
        fields_ = fields

    obid_idx_map = {o: i for i, o in enumerate(obid_level)}
    obid_label = np.empty(line_no, dtype=object)
    dt_label = np.empty(line_no, dtype=object)
    arr = np.full((line_no, len(fields_)), np.nan)
    r_map_fields = {f: i for i, f in enumerate(fields_)}

    dt_arr_sorted = np.array(sorted(dt_set))
    dt_level = convert_dt(dt_arr_sorted)

    for i, (obid, d) in enumerate(data):
        if dt_name not in d:
            continue
        dts = d[dt_name]
        slice_ = obid_slice_map[(i, obid)]
        for f, value in d.items():
            if f == dt_name:
                dt_label[slice_] = dt_arr_sorted.searchsorted(dts, side='left')
            else:
                arr[slice_, r_map_fields[f]] = value

        obid_label[slice_] = obid_idx_map[obid]

        trading_date = to_datetime(get_current_trading_date(int17_to_datetime(dts[-1])))
        if "trading_date" in r_map_fields:
            trading_date_int = date_to_int8(trading_date)
            arr[slice_, r_map_fields["trading_date"]] = trading_date_int

        daybar = daybars.get(obid)
        if daybar is not None:
            try:
                last = daybar.loc[trading_date]
            except KeyError:
                continue
            day_open = last["open"]
            if "open" in r_map_fields:
                arr[slice_, r_map_fields["open"]] = [day_open if v > 0 else 0.0 for v in d["volume"]]
            if "prev_close" in r_map_fields:
                _prev_close = arr[slice_, r_map_fields["prev_close"]][0]
                if _prev_close != _prev_close:
                    arr[slice_, r_map_fields["prev_close"]] = last["prev_close"]
            if instruments(obid).type in ("ETF", "LOF"):
                if "prev_iopv" in r_map_fields:
                    arr[slice_, r_map_fields["prev_iopv"]] = last["prev_iopv"]

            if instruments(obid).type in ("CS", "ETF", "LOF", "Future", "Spot"):
                if "limit_up" in r_map_fields:
                    arr[slice_, r_map_fields["limit_up"]] = last["limit_up"]
                if "limit_down" in r_map_fields:
                    arr[slice_, r_map_fields["limit_down"]] = last["limit_down"]

            if instruments(obid).type in ("Future", "Option", "Spot"):
                if "prev_settlement" in r_map_fields:
                    arr[slice_, r_map_fields["prev_settlement"]] = last["prev_settlement"]
                if "change_rate" in r_map_fields:
                    arr[slice_, r_map_fields["change_rate"]] = arr[slice_, r_map_fields["last"]] / last[
                        "prev_settlement"] - 1
            elif "change_rate" in r_map_fields:
                arr[slice_, r_map_fields["change_rate"]] = arr[slice_, r_map_fields["last"]] / last["prev_close"] - 1

    try:
        func_is_singletz = getattr(pd._libs.lib, 'is_datetime_with_singletz_array')
        setattr(pd._libs.lib, 'is_datetime_with_singletz_array', lambda *args: True)
    except AttributeError:
        func_is_singletz = None
    multi_idx = pd.MultiIndex(
        [obid_level, dt_level],
        [obid_label, dt_label],
        names=('order_book_id', dt_name)
    )
    df = pd.DataFrame(data=arr, index=multi_idx, columns=fields_)
    if "trading_date" in r_map_fields:
        df["trading_date"] = df["trading_date"].astype(int).apply(int8_to_datetime)
    if func_is_singletz is not None:
        setattr(pd._libs.lib, 'is_datetime_with_singletz_array', func_is_singletz)
    return df[fields]


def get_history_tick(order_book_ids, start_date, end_date, gtw_fields, columns, market):
    data = get_client().execute("get_tickbar", order_book_ids, start_date, end_date, gtw_fields, market=market)
    data = [(obid, {k: np.frombuffer(*v) for k, v in d.items()}) for obid, d in data]
    history_df = convert_history_tick_to_multi_df(data, "datetime", columns, int17_to_datetime_v)

    if isinstance(history_df, pd.DataFrame) and 'iopv' in history_df.columns:
        history_df['iopv'] = history_df['iopv'].replace(0.0, np.nan)
    return history_df


def get_tick_price_multi_df(order_book_ids, start_date, end_date, fields, market):
    ins_list = ensure_instruments(order_book_ids)
    order_book_ids = [ins.order_book_id for ins in ins_list]
    types = {ins.type for ins in ins_list}

    start_date, end_date = ensure_date_range(start_date, end_date, datetime.timedelta(days=3))
    if "Future" in types or "Option" in types or "Spot" in types:
        base_fields = FUTURE_TICK_FIELDS
        base_columns = FUTURE_TICK_COLUMNS

    elif 'ETF' in types or 'LOF' in types:
        base_fields = FUND_TICK_FIELDS
        base_columns = FUND_TICK_COLUMNS

    else:
        base_fields = EQUITIES_TICK_FIELDS
        base_columns = EQUITIES_TICK_COLUMNS

    if fields:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, base_fields, "fields")
        columns = [f for f in base_columns if f in fields]
    else:
        fields = base_fields
        columns = base_columns

    gtw_fields = set(fields) | {"date", "time"}
    if set(fields) & RELATED_DABAR_FIELDS:
        gtw_fields.update({"volume", "last"})

    history_df = get_history_tick(order_book_ids, start_date, end_date, list(gtw_fields), columns, market)
    history_latest_date = 0 if history_df is None else date_to_int8(get_current_trading_date(
        history_df.index.get_level_values(1).max()))
    today = today_int()
    next_trading_date = date_to_int8(get_next_trading_date(today, market=market))
    if history_latest_date >= end_date or start_date > next_trading_date or end_date < today:
        return history_df

    if end_date >= next_trading_date and (start_date > today or history_latest_date >= today):
        live_date = next_trading_date
    else:
        live_date = today
    if history_latest_date >= live_date:
        return history_df
    live_dfs = []
    for ins in ins_list:
        try:
            live_df = get_ticks(ins.order_book_id, start_date=live_date, end_date=live_date, expect_df=True,
                                market=market)
        except (PermissionDenied, MarketNotSupportError):
            pass
        else:
            if live_df is None:
                continue
            if "trading_date" not in live_df.columns:
                live_df["trading_date"] = int8_to_datetime(live_date)
            else:
                live_df["trading_date"] = live_df["trading_date"].apply(to_datetime)
            if ins.type in ("Future", "Option", "Spot"):
                live_df["change_rate"] = live_df["last"] / live_df["prev_settlement"] - 1
            else:
                live_df["change_rate"] = live_df["last"] / live_df["prev_close"] - 1
            live_df = live_df.reindex(columns=columns)
            live_dfs.append(live_df)

    if not live_dfs:
        return history_df

    if history_df is None:
        return pd.concat(live_dfs)
    return pd.concat([history_df] + live_dfs)


def _convert_int_to_datetime(date_int, time_int):
    return date_int * 1000000000 + time_int


def get_current_trading_date(dt):
    if 7 <= dt.hour < 18:
        return datetime.datetime(year=dt.year, month=dt.month, day=dt.day)
    return get_next_trading_date(dt - datetime.timedelta(hours=4))


def _add_minbar_dominant_id(df, dominant):
    if 'trading_date' in df.columns:
        dominant.index = dominant.index.map(lambda x: float(x.year * 10000 + x.month * 100 + x.day))
        date_dominant_map = dominant.to_dict()
        df['dominant_id'] = df['trading_date'].map(date_dominant_map)
    else:
        date_dominant_map = dominant.to_dict()

        def _set_dominant(dt):
            trading_date = pd.Timestamp(get_current_trading_date(dt))
            return date_dominant_map[trading_date]

        df['dominant_id'] = df.index.map(_set_dominant)
    return df



# -*- coding: utf-8 -*-
import warnings
import numpy as np
import pandas as pd
from rqdatac.services.detail.resample_helper import resample_week_df


from rqdatac.services.get_price import (
    _ensure_fields,
    get_current_trading_date
)
from rqdatac.utils import (
    int14_to_datetime_v,
    int8_to_datetime_v,
    today_int,
    date_to_int8,
)
from rqdatac.client import get_client
from rqdatac.services.future import get_dominant_future
from rqdatac.services.stock_status import is_suspended
from rqdatac.share.errors import PermissionDenied, MarketNotSupportError

DAYBAR_FIELDS = {
    "future": ["settlement", "prev_settlement", "open_interest", "limit_up", "limit_down"],
    "common": ["open", "close", "high", "low", "total_turnover", "volume"],
    "stock": ["limit_up", "limit_down", "num_trades"],
    "fund": ["limit_up", "limit_down", "num_trades", "iopv"],
    "spot": ["settlement", "prev_settlement", "open_interest", "limit_up", "limit_down"],
    "option": ["open_interest", "strike_price", "contract_multiplier", "prev_settlement", "settlement", "limit_up",
               "limit_down"],
    "convertible": ["num_trades"],
    "index": [],
    "repo": ["num_trades"],
}

WEEKBAR_FIELDS = {
    "future": ["settlement", "prev_settlement", "open_interest"],
    "common": ["open", "close", "high", "low", "total_turnover", "volume"],
    "stock": ["num_trades"],
    "fund": ["num_trades", "iopv"],
    "spot": ["settlement", "prev_settlement", "open_interest"],
    "option": ["open_interest", "strike_price", "contract_multiplier", "settlement"],
    "convertible": ["num_trades"],
    "index": [],
    "repo": ["num_trades"],
}

MINBAR_FIELDS = {
    "future": ["trading_date", "open_interest"],
    "common": ["open", "close", "high", "low", "total_turnover", "volume"],
    "stock": ["num_trades"],
    "fund": ["num_trades", "iopv"],
    "spot": ["trading_date", "open_interest"],
    "option": ["trading_date", "open_interest"],
    "convertible": [],
    "index": [],
    "repo": [],
}


ZERO_FILL_FIELDS = frozenset({"total_turnover", "open_interest", "volume"})

SPOT_DIRECTION_MAP = {0: "null", 1: "多支付空", 2: "空支付多", 3: "交收平衡"}


def get_price_df(
        order_book_ids,
        start_date,
        end_date,
        frequency,
        duration,
        fields,
        adjust_type,
        skip_suspended,
        stocks,
        funds,
        indexes,
        futures,
        futures888,
        spots,
        options,
        convertibles,
        repos,
        market
):
    if frequency == "d":
        fields, has_dominant_id = _ensure_fields(fields, DAYBAR_FIELDS, stocks, funds, futures, futures888, spots, options, convertibles, indexes, repos)
        pf, obid_slice_map = get_daybar(order_book_ids, start_date, end_date, fields, duration, market)
        if pf is None:
            return
    else:
        fields, has_dominant_id = _ensure_fields(fields, MINBAR_FIELDS, stocks, funds, futures, futures888, spots, options, convertibles, indexes, repos)
        history_permission_denied, today_permission_denied = False, False
        try:
            pf, obid_slice_map = get_minbar(order_book_ids, start_date, end_date, fields, duration, market)
        except (PermissionDenied, MarketNotSupportError):
            pf = obid_slice_map = None
            history_permission_denied = True
        history_latest_day = 0 if pf is None else date_to_int8(pf.index.get_level_values(1).max())
        if history_latest_day < end_date and end_date >= today_int():
            try:
                today_pf, today_obid_slice_map = get_today_minbar(order_book_ids, fields, duration, market)
            except (PermissionDenied, MarketNotSupportError):
                today_pf = None
                today_permission_denied = True
            if today_pf is None:
                today_pf_latest_day = 0
            else:
                today_pf_latest_day = date_to_int8(get_current_trading_date(today_pf.index.get_level_values(1).max()))
            if today_pf_latest_day > history_latest_day and today_pf_latest_day >= start_date:
                if history_latest_day == 0:
                    pf = today_pf
                    obid_slice_map = today_obid_slice_map
                else:
                    pf = pd.concat([pf, today_pf])
                    line_no, obid_slice_map = 0, {}
                    obids, counts = np.unique(pf.index.get_level_values(0), return_counts=True)
                    for obid, ct in zip(obids, counts):
                        obid_slice_map[obid] = slice(line_no, line_no + ct, None)
                        line_no += ct
        if pf is None:
            if history_permission_denied and today_permission_denied:
                raise PermissionDenied("Not permit to get minbar price ")
            elif history_permission_denied:
                warnings.warn("Not permit to get history minbar price")
            elif today_permission_denied:
                warnings.warn("Not permit to get realtime minbar price")
            return

    result = _adjust_pf(
        pf,
        order_book_ids,
        stocks,
        funds,
        convertibles,
        futures888,
        start_date,
        end_date,
        has_dominant_id,
        adjust_type,
        skip_suspended,
        obid_slice_map,
        market,
    )
    return result


def get_daybar(order_book_ids, start_date, end_date, fields, duration, market):
    data = get_client().execute(
        "get_daybar_v", order_book_ids, start_date, end_date, fields, duration, market
    )
    data = [(obid, {k: np.frombuffer(*v) for k, v in d.items()}) for obid, d in data]
    return convert_bar_to_multi_df(data, 'date', fields, int8_to_datetime_v)


def get_future_indx_daybar(order_book_ids, start_date, end_date, fields, duration=1, market="cn"):
    data = get_client().execute(
        "futures.get_future_indx_daybar_v", order_book_ids, start_date, end_date, fields, duration, market
    )
    data = [(obid, {k: np.frombuffer(*v) for k, v in d.items()}) for obid, d in data]
    multi_df, _ = convert_bar_to_multi_df(data, 'date', fields, int8_to_datetime_v)
    if multi_df is not None:
        return multi_df


def get_minbar(order_book_ids, start_date, end_date, fields, duration, market):
    data = get_client().execute(
        "get_minbar_v", order_book_ids, start_date, end_date, fields, duration, market
    )
    data = [(obid, {k: np.frombuffer(*v) for k, v in d.items()}) for obid, d in data]
    return convert_bar_to_multi_df(data, "datetime", fields, int14_to_datetime_v)


def get_today_minbar(order_book_ids, fields, duration, market="cn"):
    data = get_client().execute("get_today_minbar", order_book_ids, fields, duration, market)
    return convert_bar_to_multi_df(data, "datetime", fields, int14_to_datetime_v)


def convert_bar_to_multi_df(data, dt_name, fields, convert_dt):
    line_no = 0
    dt_set = set()
    obid_level = []
    obid_slice_map = {}
    for obid, d in data:
        dts = d[dt_name]
        dts_len = len(dts)
        if dts_len == 0:
            continue
        obid_slice_map[obid] = slice(line_no, line_no + dts_len, None)
        dt_set.update(dts)
        line_no += dts_len

        obid_level.append(obid)

    if line_no == 0:
        return None, obid_slice_map

    obid_idx_map = {o: i for i, o in enumerate(obid_level)}
    obid_label = np.empty(line_no, dtype=object)
    dt_label = np.empty(line_no, dtype=object)
    arr = np.full((line_no, len(fields)), np.nan)
    r_map_fields = {f: i for i, f in enumerate(fields)}
    for f in ZERO_FILL_FIELDS:
        if f in fields:
            arr[:, r_map_fields[f]] = 0

    dt_arr_sorted = np.array(sorted(dt_set), dtype=np.int64)
    dt_level = convert_dt(dt_arr_sorted)

    for obid, d in data:
        dts = d[dt_name]
        if len(dts) == 0:
            continue
        slice_ = obid_slice_map[obid]
        for f, value in d.items():
            if f == dt_name:
                dt_label[slice_] = dt_arr_sorted.searchsorted(dts, side='left')
            else:
                arr[slice_, r_map_fields[f]] = value
        obid_label[slice_] = [obid_idx_map[obid]] * len(dts)

    try:
        func_is_singletz = getattr(pd._libs.lib, 'is_datetime_with_singletz_array')
        setattr(pd._libs.lib, 'is_datetime_with_singletz_array', lambda *args: True)
    except AttributeError:
        func_is_singletz = None
    multi_idx = pd.MultiIndex(
        [obid_level, dt_level],
        [obid_label, dt_label],
        names=('order_book_id', dt_name)
    )
    if func_is_singletz is not None:
        setattr(pd._libs.lib, 'is_datetime_with_singletz_array', func_is_singletz)

    df = pd.DataFrame(data=arr, index=multi_idx, columns=fields)
    return df, obid_slice_map


def _adjust_pf(
        pf,
        order_book_ids,
        stocks,
        funds,
        convertibles,
        futures888,
        start_date,
        end_date,
        has_dominant_id,
        adjust_type,
        skip_suspended,
        obid_slice_map,
        market,
):
    adjust = (stocks or funds) and adjust_type in {"pre", "post", "pre_volume", "post_volume"}
    if adjust:
        from rqdatac.services.detail.adjust_price import adjust_price_multi_df
        adjust_price_multi_df(pf, stocks + funds, adjust_type, obid_slice_map, market)
    if has_dominant_id:
        # 1.全为非正常合约 2.有期货类型合约并且指定dominant_id字段
        # 只有满足其中一种才在返回字段中增加dominant_id
        add_dominant_id(pf, futures888, obid_slice_map)
    if skip_suspended and len(order_book_ids) == 1 and (stocks or convertibles):
        pf = filter_suspended(pf, order_book_ids[0], start_date, end_date, len(convertibles) > 0, market)

    if "trading_date" in pf:

        def convert_to_timestamp(v):
            if np.isnan(v):
                return pd.NaT
            return pd.Timestamp(str(int(v)))

        if hasattr(pf.trading_date, "applymap"):
            pf.trading_date = pf.trading_date.applymap(convert_to_timestamp)
        else:
            pf.trading_date = pf.trading_date.apply(convert_to_timestamp)

    if "settlement_direction" in pf:

        def convert_direction(key):
            if np.isnan(key):
                return key
            return SPOT_DIRECTION_MAP[key]

        if hasattr(pf.settlement_direction, "applymap"):
            pf.settlement_direction = pf.settlement_direction.applymap(convert_direction)
        else:
            pf.settlement_direction = pf.settlement_direction.apply(convert_direction)

    return pf


def add_dominant_id(result, futures888, obid_slice_map):
    from rqdatac.services.calendar import get_next_trading_date
    def _may_shift_date(d):
        # 夜盘的开始时间是晚上9点, 这时候对应的 dominant_id 应该是下一个交易日的 dominant_id
        if d.hour >= 21:
            return pd.Timestamp(get_next_trading_date(d))
        return d

    for order_book_id, underlying in futures888.items():
        if order_book_id in obid_slice_map:
            slice_ = obid_slice_map[order_book_id]
            dts = result.index.get_level_values(1)[slice_].map(_may_shift_date)
            dominants = get_dominant_future(
                underlying, dts[0].date(), dts[-1].date())
            if dominants is not None:
                result.at[slice_, "dominant_id"] = np.take(
                    dominants.values, dominants.index.searchsorted(dts, side="right") - 1)


def filter_suspended(ret, order_book_id, start_date, end_date, is_convertible, market):
    if is_convertible:
        from rqdatac.services.convertible import is_suspended as is_convertible_suspend
        s = is_convertible_suspend(order_book_id, start_date, end_date)
    else:
        s = is_suspended(order_book_id, start_date, end_date, market)
    ret_date_index = ret.index.get_level_values(1)
    index = s.index.union(ret_date_index)
    s = s.reindex(index)
    s = s.fillna(method="ffill")
    s = s.loc[ret_date_index]
    s = s[order_book_id] == False
    return ret[s.values]


def get_week_df(order_book_ids, start_date, end_date, fields, adjust_type, market, stocks, funds, indexes, futures,
                futures888, spots, options, convertibles, repos):
    fields, has_dominant_id = _ensure_fields(fields, WEEKBAR_FIELDS, stocks, funds, futures, futures888, spots,
                                             options, convertibles, indexes, repos)
    has_volume_field = 'volume' in fields
    if not has_volume_field:
        fields.append('volume')
    df = get_price_df(
        order_book_ids, start_date, end_date, 'd', 1, fields, adjust_type, False,
        stocks, funds, indexes, futures, futures888, spots, options, convertibles, repos, market
    )
    if df is None:
        return
    res = resample_week_df(df, fields)
    if not has_volume_field:
        res.drop(columns=['volume'], inplace=True)
    return res



# coding: utf-8
from sqlalchemy import Numeric, Column
from .fundamental_base_sql import FundamentalBase


class AnaStkValIdx(FundamentalBase):
    pe_ratio = Column(Numeric(18, 4))
    pcf_ratio = Column(Numeric(18, 4))
    pb_ratio = Column(Numeric(18, 4))
    market_cap = Column(Numeric(21, 4))
    market_cap_2 = Column(Numeric(21, 4))
    a_share_market_val = Column(Numeric(21, 4))
    a_share_market_val_2 = Column(Numeric(21, 4))
    val_of_stk_right = Column(Numeric(21, 4))
    dividend_yield = Column(Numeric(18, 4))
    pe_ratio_1 = Column(Numeric(18, 4))
    pe_ratio_2 = Column(Numeric(18, 4))
    peg_ratio = Column(Numeric(18, 4))
    pcf_ratio_1 = Column(Numeric(18, 4))
    pcf_ratio_2 = Column(Numeric(18, 4))
    pcf_ratio_3 = Column(Numeric(18, 4))
    ps_ratio = Column(Numeric(18, 4))
    enterprise_value = Column(Numeric(19, 4))
    enterprise_value_2 = Column(Numeric(19, 4))



# -*- coding: utf-8 -*-
import datetime
import six
import warnings
from itertools import islice
from functools import reduce

import pandas as pd
import numpy as np
import math
from dateutil.relativedelta import relativedelta
from sqlalchemy import Column
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.elements import UnaryExpression
from sqlalchemy.dialects import mysql
from sqlalchemy.ext.declarative import DeclarativeMeta
from pymysql.converters import conversions, escape_item, encoders

from rqdatac.client import get_client
from rqdatac.services.orm.balance_sheet_sql import StkBalaGen
from rqdatac.services.orm.cash_flow_sql import StkCashGen
from rqdatac.services.orm.financial_indicator_sql import AnaStkFinIdx
from rqdatac.services.orm.fundamental_base_sql import FundamentalBase
from rqdatac.services.orm.eod_derivative_indicator_sql import AnaStkValIdx
from rqdatac.services.orm.income_statement_sql import StkIncomeGen
from rqdatac.services.orm.ttm_sql import (
    CashFlowStatementTTM,
    IncomeStatementTTM,
    FinancialIndicatorTTM,
)
from rqdatac.services.orm.pit_financials import (
    BalanceSheet,
    CashFlowsStatement,
    IncomeStatement,
    MainData,
    TTM,
    PIT_TABLES,
    PIT_FIELDS,
    PIT_BASIC_FIELDS,
    PIT_BASIC_FIELDS_SET
)

from rqdatac.services.orm.pit_financials_ex import FIELDS_LIST_EX
from rqdatac.share.errors import MarketNotSupportError
from rqdatac.services.calendar import get_previous_trading_date
from rqdatac.utils import int8_to_date, to_date_int, to_datetime, is_panel_removed
from rqdatac.validators import (
    ensure_list_of_string,
    ensure_string,
    check_items_in_container,
    ensure_date_int,
    ensure_order_book_id,
    raise_for_no_panel,
    check_quarter,
    ensure_date_or_today_int,
    quarter_string_to_date,
    ensure_order_book_ids,
)
from rqdatac.decorators import export_as_api


@export_as_api(name="financials")
@export_as_api
class Financials:
    stockcode = FundamentalBase.stockcode
    announce_date = FundamentalBase.announce_date
    income_statement = StkIncomeGen
    balance_sheet = StkBalaGen
    cash_flow_statement = StkCashGen
    financial_indicator = AnaStkFinIdx
    cash_flow_statement_TTM = CashFlowStatementTTM
    income_statement_TTM = IncomeStatementTTM
    financial_indicator_TTM = FinancialIndicatorTTM


@export_as_api
def get_financials(query, quarter, interval=None, expect_df=True, market="cn"):
    """获取季度报表数据

    :param query: SQLAlchemy Query对象
    :param quarter: 季度, 如: '2016q1'
    :param interval: 2y', '4q', 与 get_fundamentals 类似，但只接受 y(year) 和 q(quarter)
        (Default value = None)
    :param market: 市场 (Default value = "cn")
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :returns: 如果返回结果中股票代码和查询指标为单个值, 返回Series;
    如果返回结果中股票代码或查询指标中有且只有一个为单个值, 返回pandas.DataFrame; 否则返回pandas.Panel

    """

    msg = "'get_financials' is deprecated, and will be removed soon. use get_pit_financials_ex instead."
    warnings.warn(msg, stacklevel=2)

    sql, quarters = _parse_arguments(query, quarter, interval)
    records = get_client().execute("get_financial", sql, market=market)
    return parse_results(records, quarters, expect_df)


def classify_fields(fields):
    if isinstance(fields, six.string_types) or fields in (
            BalanceSheet, CashFlowsStatement, IncomeStatement, TTM, MainData):
        fields = [fields]
    result = [
        # table_name fields
        ('balance_sheet_new', []),
        ('cash_flow_statement_new', []),
        ('income_statement_new', []),
        ('ttm_new', []),
        ('main_data', []),
        ('rnd_expenditure_sheet', []),
    ]

    for f in fields:
        if f in PIT_BASIC_FIELDS_SET:
            continue
        elif f in PIT_FIELDS[0]:
            result[0][1].append(f)
        elif f in PIT_FIELDS[1]:
            result[1][1].append(f)
        elif f in PIT_FIELDS[2]:
            result[2][1].append(f)
        elif f in PIT_FIELDS[3]:
            result[3][1].append(f)
        elif f in PIT_FIELDS[4]:
            result[4][1].append(f)
        elif f in PIT_FIELDS[5]:
            result[5][1].append(f)
        elif f in (BalanceSheet, CashFlowsStatement, IncomeStatement, TTM, MainData):
            index = PIT_TABLES.index(f.__name__)
            result[index][1].extend(PIT_FIELDS[index])
        else:
            raise ValueError('invalid field: {}'.format(f))
    return result


@export_as_api(name="pit_financials")
@export_as_api
class PitFinancials:
    balance_sheet = BalanceSheet
    cash_flows_statement = CashFlowsStatement
    income_statement = IncomeStatement
    ttm = TTM
    main_data = MainData


ENTERPRISE_TYPE_MAP = {
    13: "business_bank",
    31: "securities_firms",
    33: "trust",
    35: "insurance_company",
    39: "other_financial_institution",
    99: "general_enterprise",
}

INFO_TYPE_MAP = {
    10: "发行上市书",
    20: "定期报告",
    30: "业绩快报",
    50: "章程制度",
    70: "临时公告",
    90: "交易所通报",
    91: "交易所临时停(复)牌公告",
    99: "其他",
    110101: "定期报告:年度报告",
    110102: "定期报告:半年度报告",
    110103: "定期报告:第一季报",
    110104: "定期报告:第三季报",
    110105: "定期报告:审计报告",
    110106: "定期报告:第二季报",
    110107: "定期报告:第四季报",
    110108: "定期报告:第五季报",
    110109: "定期报告:第二季报（更正后）",
    110110: "定期报告:第四季报（更正后）",
    110111: "定期报告:第五季报（更正后）",
    110201: "定期报告:年度报告(关联方)",
    110202: "定期报告:半年度报告(关联方)",
    110203: "定期报告:第一季报(关联方)",
    110204: "定期报告:第三季报(关联方)",
    120101: "临时公告:审计报告(更正后)",
    120102: "临时公告:年度报告(更正后)",
    120103: "临时公告:半年度报告(更正后)",
    120104: "临时公告:第一季报(更正后)",
    120105: "临时公告:第三季报(更正后)",
    120106: "临时公告:公开转让说明书(更正后)",
    120107: "临时公告:业绩快报",
    120108: "临时公告:业绩快报(更正后)",
    120201: "临时公告:跟踪评级报告",
    120202: "临时公告:同业存单发行计划",
    120203: "临时公告:比较式财务报表",
    120204: "临时公告:关联方",
    120205: "临时公告:其他",
    120206: "临时公告:前期差错更正",
    120207: "临时公告:第一季度报告",
    120208: "临时公告:第二季度报告",
    120209: "临时公告:第三季度报告",
    120210: "临时公告:第四季度报告",
    120211: "临时公告：年度报告",
    130101: "发行上市书:募集说明书",
    130102: "发行上市书:招股说明书(申报稿)",
    130103: "发行上市书:招股意向书",
    130104: "发行上市书:上市公告书",
    130105: "发行上市书:审阅报告",
    130106: "发行上市书:招股说明书",
    130107: "发行上市书:公开转让说明书",
    130108: "发行上市书:发行公告",
    130109: "发行上市书:审计报告",
    130110: "发行上市书:关联方",
    130111: "发行上市书:其他",
    140101: "发行披露文件:第一季报",
    140102: "发行披露文件:半年度报告",
    140103: "发行披露文件：第三季报",
    140104: "发行披露文件：审计报告",
    140105: "发行披露文件：募集说明书",
    140106: "发行披露文件：跟踪评级报告"
}


@export_as_api
def get_pit_financials(fields, quarter, interval=None, order_book_ids=None,
                       if_adjusted='all', max_info_date=None, market='cn'):
    """
    获取pit季度报表数据
    :param fields: 财务指标 or 财务指标 list
    :param quarter: 季度, 如: '2016q1'
    :param interval: '2y', '4q', 默认只返回当季
    :param order_book_ids: 股票列表, 默认为None返回所有股票数据
    :param if_adjusted: 是否调整
        0: 每个order_book_id每个季度返回一条发布日期最新的未调整的数据
        1: 每个order_book_id每个季度返回一条发布日期最新的调整的数据
        'ignore': 每个order_book_id每个季度返回一条发布日期最新的数据
        'all': 返回所有数据,无论发布日期新旧和是否调整
    :param max_info_date: 指定最大发布日期, 如20180430，则所取数据的发布日期均不大于20180430，默认为None
    :return: pandas.DataFrame or None
    """

    msg = "'get_pit_financials' is deprecated, and will be removed soon. use get_pit_financials_ex instead."
    warnings.warn(msg, stacklevel=2)

    if if_adjusted not in [0, 1, '0', '1', 'all', 'ignore']:
        raise ValueError("if_adjusted should in [0, 1, 'all', 'ignore']")

    if order_book_ids is not None:
        order_book_ids = ensure_list_of_string(order_book_ids)
    quarters = get_quarters(quarter, interval)
    quarter_dates = [str(quarter_to_date(y, q)) for y, q in quarters]
    result = []
    for i, (table_name, fields) in enumerate(classify_fields(fields)):
        if not fields:
            continue

        if i == 3:  # ttm, exclude ['info_type', 'is_complete', 'enterprise_type']
            extra_fields = PIT_BASIC_FIELDS[0:-3]
        elif i == 4:  # main_data, exclude ['is_complete', 'enterprise_type']
            extra_fields = PIT_BASIC_FIELDS[0:-2]
        elif i == 5:
            extra_fields = ("order_book_id", "end_date", "if_adjusted", "info_date", "info_type")
        else:
            extra_fields = PIT_BASIC_FIELDS
        fields.extend(extra_fields)

        sql = """SELECT {}
        FROM {}
        """.format(', '.join(fields), table_name)
        where = ["end_date IN ('{}')".format("', '".join(quarter_dates))]
        if if_adjusted not in ['all', 'ignore']:
            where.append('if_adjusted {} 2'.format('=' if if_adjusted in (0, '0') else '<>'))
        if order_book_ids is not None:
            where.append("order_book_id IN ('{}')".format("', '".join(order_book_ids)))
        if max_info_date is not None:
            where.append("info_date <= '{}'".format(max_info_date))

        sql += """WHERE {}""".format(' AND '.join(where))
        records = get_client().execute("get_pit_financials", sql, market=market)
        if records:
            df = pd.DataFrame(records)
            df['if_adjusted'] = df['if_adjusted'].apply(lambda x: 0 if x == 2 else 1)

            if 'accounting_standards' in df.columns:
                df['accounting_standards'] = df['accounting_standards'].apply(lambda x: 1 if x == 1 else 0)
            if 'if_complete' in df.columns:
                df['if_complete'] = df['if_complete'].apply(lambda x: 1 if x == 1 else 0)

            if if_adjusted != 'all':
                if if_adjusted == 'ignore':
                    subset = ['order_book_id', 'end_date']
                else:
                    subset = ['order_book_id', 'end_date', 'if_adjusted']
                df.sort_values('info_date', inplace=True)
                df.drop_duplicates(subset=subset, keep='last', inplace=True)
            df.fillna(np.inf, inplace=True)
            result.append(df)
    if not result:
        return
    if len(result) == 1:
        result = result[0]
    elif len(result) > 1:
        result = reduce(
            lambda left, right: pd.merge(
                left, right, how='outer', on=[f for f in left.columns if f in right.columns]
            ), result)

    result.sort_values(by=['order_book_id', 'end_date', 'info_date'], inplace=True)
    result.set_index(['order_book_id', 'end_date'], inplace=True)

    base_columns = ['info_date', 'if_adjusted']
    if 'accounting_standards' in result.columns:
        base_columns.append('accounting_standards')
    if 'if_complete' in result.columns:
        result.rename(columns={'if_complete': 'is_complete'}, inplace=True)
        result['enterprise_type'] = result['enterprise_type'].map(ENTERPRISE_TYPE_MAP)
        base_columns.extend(['is_complete', 'enterprise_type'])
    if 'info_type' in result.columns:
        result['info_type'] = result['info_type'].map(INFO_TYPE_MAP)
        base_columns.extend(['info_type'])

    return result[base_columns + [c for c in result.columns if c not in base_columns]]


@export_as_api
def get_pit_financials_ex(order_book_ids, fields, start_quarter, end_quarter,
                          date=None, statements='latest', market='cn'):
    """
        获取股票财务数据(Point In Time)
    :param order_book_ids: 股票合约代码列表
    :param fields: 指定返回财报字段
    :param start_quarter: 财报季度 - 起始，如 2020q1
    :param end_quarter: 财报季度 - 截止
    :param date: 财报发布日期，默认为当前日期, 如 '2020-01-01' | '20200101'
    :param statements: 可选 latest/all, 默认为 latest
            latest: 仅返回在date时点所能观察到的最新数据；
            all：返回在date时点所能观察到的所有版本，从第一次发布直到观察时点的所有修改。
    :param market: 股票市场范围
    :return:
    """
    fields = ensure_list_of_string(fields, 'fields')
    check_items_in_container(fields, FIELDS_LIST_EX, "fields")
    fields.extend(['order_book_id', 'info_date', 'end_date', 'if_adjusted'])
    fields = list(set(fields))
    fields[fields.index("info_date")], fields[0] = fields[0], fields[fields.index("info_date")]

    check_quarter(start_quarter, 'start_quarter')
    start_quarter_int = ensure_date_int(quarter_string_to_date(start_quarter))

    check_quarter(end_quarter, 'end_quarter')
    end_quarter_int = ensure_date_int(quarter_string_to_date(end_quarter))

    if start_quarter > end_quarter:
        raise ValueError(
            'invalid quarter range: [{!r}, {!r}]'.format(
                start_quarter, end_quarter))

    date = ensure_date_or_today_int(date)

    order_book_ids = ensure_list_of_string(order_book_ids, 'order_book_ids')

    if statements not in ['all', 'latest']:
        raise ValueError("invalid statements , got {!r}".format(statements))

    pit_financial_df = pd.DataFrame(
        get_client().execute("get_pit_financials_ex", order_book_ids, fields, start_quarter_int, end_quarter_int, date,
                             statements, market))
    if pit_financial_df.empty:
        return
    pit_financial_df = pit_financial_df.reindex(columns=fields)
    pit_financial_df.sort_values(['order_book_id', 'end_date', 'info_date'])
    pit_financial_df["end_date"] = pit_financial_df["end_date"].apply(
        lambda d: "{}q{}".format(d.year, math.ceil(d.month / 3)))
    pit_financial_df.rename(columns={"end_date": "quarter"}, inplace=True)
    pit_financial_df.set_index(['order_book_id', 'quarter'], inplace=True)
    pit_financial_df['if_adjusted'] = pit_financial_df['if_adjusted'].map(lambda x: 1 if x == 1 else 0).astype(int)
    pit_financial_df.sort_index(inplace=True)
    return pit_financial_df


@export_as_api
def get_fundamentals(query, entry_date, interval=None, report_quarter=False, expect_df=True, market="cn"):
    """获取财务数据

    :param query: query 对象
    :param entry_date: 日期
    :param interval:  (Default value = None)
    :param report_quarter:  (Default value = False)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :param market:  (Default value = "cn")

    """

    msg = "'get_fundamentals' is deprecated, and will be removed soon. use get_factor instead."
    warnings.warn(msg, stacklevel=2)

    if market != "cn":
        raise MarketNotSupportError("don't support market {} yet.", market)

    if not isinstance(query, Query):
        raise ValueError("a sqlalchemy's Query object expected: {}".format(type(query)))

    raise_for_no_panel(expect_df)
    entry_date = to_datetime(entry_date)
    delta = 0
    duration = 0
    if interval is not None:
        if not isinstance(interval, str):
            raise ValueError(
                "invalid interval: {} should be a string like 1d, 5y, 3m, 2q".format(interval)
            )
        if interval[-1] not in __TIME_DELTA_MAP__:
            raise ValueError(
                "invalid interval: {}, interval unit should be d(day), "
                "m(month), q(quarter) or y(year)".format(interval)
            )
        delta = __TIME_DELTA_MAP__[interval[-1]]

        try:
            duration = int(interval[:-1])
        except ValueError:
            raise ValueError(
                "invalid interval: {}, should be a string like 1d, 5y, 3m, 2q".format(interval)
            )

    trading_dates = [get_previous_trading_date(entry_date + __TIME_DELTA_MAP__["d"], market=market)]
    if duration > 0:
        current_date = trading_dates[0]
        one_day = __TIME_DELTA_MAP__["d"]
        for i in range(duration - 1):
            current_date = get_previous_trading_date(current_date - delta + one_day, market=market)
            trading_dates.append(current_date)

    query = _unsafe_apply_query_filter(query, trading_dates)
    sql = _compile_query(query)
    records = get_client().execute("get_fundamentals", sql, market=market)

    if not records:
        warnings.warn("No record found")
        return None

    base_fields = ["STOCKCODE", "TRADEDATE", "RPT_YEAR", "RPT_QUARTER"]
    field_names = base_fields + list(set(records[0].keys()) - set(base_fields))
    items = ["report_quarter"] + field_names[4:] if report_quarter else field_names[4:]

    if expect_df:
        df = pd.DataFrame(records)
        df.rename(columns={"STOCKCODE": "order_book_id", "TRADEDATE": "date"}, inplace=True)
        df["report_quarter"] = df["RPT_YEAR"].map(str) + "q" + df["RPT_QUARTER"].map(str)
        df.sort_values(["order_book_id", "date"], ascending=[True, False], inplace=True)
        df.set_index(["order_book_id", "date"], inplace=True)
        for item in items:
            if item != "report_quarter":
                df[item] = df[item].astype(np.float64)
        return df[items]

    # 只有一个查询日期时, 保持顺序
    if len(trading_dates) > 1:
        stocks = list(set([r[field_names[0]] for r in records]))
    else:
        stocks = [r[field_names[0]] for r in records]

    stock_index = {s: i for i, s in enumerate(stocks)}
    day_index = {d: i for i, d in enumerate(trading_dates)}

    removed_items_size = 3 if report_quarter else 4

    array = np.ndarray(
        ((len(records[0]) - removed_items_size), len(trading_dates), len(stocks)), dtype=object
    )
    array.fill(np.nan)
    for r in records:
        istock = stock_index[r[field_names[0]]]
        iday = day_index[int8_to_date(r[field_names[1]])]
        for i in range(4, len(r)):
            array[(i - removed_items_size, iday, istock)] = np.float64(r[field_names[i]])
        if report_quarter:
            array[(0, iday, istock)] = (
                np.nan
                if None in (r[field_names[2]], r[field_names[3]])
                else str(r[field_names[2]]) + "q" + str(r[field_names[3]])
            )

    trading_dates = pd.to_datetime(trading_dates)

    warnings.warn("Panel is  removed after pandas version 0.25.0."
                  "the  default value of 'expect_df' will change to True in the future.")
    return pd.Panel(data=array, items=items, major_axis=trading_dates, minor_axis=stocks)


deprecated_fields = {"data_point": "data_point", "table": "table", "comment": "comment"}


@export_as_api
def deprecated_fundamental_data(fields=None, market="cn"):
    fields = _check_deprecated_fields(fields)
    df = pd.DataFrame(get_client().execute("deprecated_fundamental_data", fields, market=market))

    if len(df) < 1:
        return None
    if len(fields) == 1:
        df = list(df[fields[0]])
    return df


@export_as_api
def current_performance(
        order_book_id, info_date=None, quarter=None, interval="1q", fields=None, market="cn"
):
    """获取A股快报

    :param order_book_id: 股票代码, 如'000001.XSHE'
    :param info_date: 发布日期, 如'20180501', 默认为最近的交易日 (Default value = None)
    :param quarter: 发布季度, 如'2018q1' (Default value = None)
    :param interval: 数据区间， 发布日期, 如'2y', '4q' (Default value = "1q")
    :param fields: str 或 list 类型. 默认为 None, 返回所有字段 (Default value = None)
    :param market: 地区代码, 如'cn' (Default value = "cn")
    :returns: pd.DataFrame

    """
    order_book_id = ensure_order_book_id(order_book_id, market=market)
    end_date = None
    if info_date:
        info_date = ensure_date_int(info_date)
    elif quarter:
        splited = quarter.lower().split("q")
        if len(quarter) != 6 or len(splited) != 2:
            raise ValueError(
                "invalid argument {}: {}, valid parameter: {}".format(
                    "quarter", quarter, "string format like '2016q1'"
                )
            )

        year, quarter = int(splited[0]), int(splited[1])
        if not 1 <= quarter <= 4:
            raise ValueError(
                "invalid argument {}: {}, valid parameter: {}".format(
                    "quarter", quarter, "quarter should be in [1, 4]"
                )
            )
        month, day = QUARTER_DATE_MAP[quarter]
        end_date = ensure_date_int(datetime.datetime(year, month, day))
    else:
        info_date = ensure_date_int(datetime.date.today())
    ensure_string(interval, "interval")
    if interval[-1] not in ("y", "q", "Y", "Q"):
        raise ValueError(
            "invalid argument {}: {}, valid parameter: {}".format(
                "interval", interval, "interval unit should be q(quarter) or y(year)"
            )
        )

    try:
        int(interval[:-1])
    except ValueError:
        raise ValueError(
            "invalid argument {}: {}, valid parameter: {}".format(
                "interval", interval, "string like 4q, 2y"
            )
        )
    interval = interval.lower()

    if fields is not None:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, PERFORMANCE_FIELDS, "fields")
    else:
        fields = PERFORMANCE_FIELDS

    data = get_client().execute(
        "current_performance", order_book_id, info_date, end_date, fields, market=market
    )
    if not data:
        return
    df = pd.DataFrame(data)
    sort_field = "info_date" if info_date else "end_date"
    df.sort_values(by=[sort_field, "mark"], ascending=[False, True], inplace=True)
    df.drop_duplicates(subset="end_date", keep="first", inplace=True)
    num = int(interval[:-1])
    unit = interval[-1]
    if unit == "y":
        latest_month = df.loc[0, "end_date"].month
        df["month"] = df.end_date.apply(lambda x: x.month)
        df = df[df.month == latest_month]
    df.reset_index(drop=True, inplace=True)
    return df.loc[: num - 1, ["end_date", "info_date"] + fields]


PERFORMANCE_FORECAST_FIELDS = [
    "forecast_type",
    "forecast_description",
    "forecast_growth_rate_floor",
    "forecast_growth_rate_ceiling",
    "forecast_earning_floor",
    "forecast_earning_ceiling",
    "forecast_np_floor",
    "forecast_np_ceiling",
    "forecast_eps_floor",
    "forecast_eps_ceiling",
    "net_profit_yoy_const_forecast",
]


@export_as_api
def performance_forecast(order_book_ids, info_date=None, end_date=None, fields=None, market="cn"):
    """获取业绩预报

    :param order_book_ids: 股票代码，如['000001.XSHE', '000002.XSHE']
    :param info_date: 信息发布日期，如'20180501'，默认为最近的交易日 (Default value = None)
    :param end_date: 业绩预计报告期，如'20180501'，默认为最近的交易日 (Default value = None)
    :param fields: str或list类型. 默认为None，返回所有字段 (Default value = None)
    :param market:  (Default value = "cn")
    :returns: pd.DataFrame

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, type='CS')
    if info_date:
        info_date = ensure_date_int(info_date)
    elif end_date:
        end_date = ensure_date_int(end_date)
    else:
        info_date = ensure_date_int(datetime.datetime.today())

    if fields:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, PERFORMANCE_FORECAST_FIELDS, "fields")
    else:
        fields = PERFORMANCE_FORECAST_FIELDS

    data = get_client().execute(
        "performance_forecast", order_book_ids, info_date, end_date, fields, market=market
    )
    if not data:
        return
    if len(order_book_ids) > 1:
        df = pd.DataFrame(data, columns=["order_book_id", "info_date", "end_date"] + fields)
        return df.set_index("order_book_id")

    return pd.DataFrame(data, columns=["info_date", "end_date"] + fields)


def _check_deprecated_fields(fields):
    try:
        if fields is None:
            fields = list(deprecated_fields.values())
        elif isinstance(fields, str):
            fields = [deprecated_fields[fields]]
        elif isinstance(fields, list):
            fields = [deprecated_fields[i] for i in fields]
        else:
            raise ValueError("fields should be string or list of strings")
    except KeyError:
        raise ValueError(
            "invalid argument fields: {!r}, valid parameter: {!r}".format(
                fields, list(deprecated_fields)
            )
        )
    return fields


PERFORMANCE_FIELDS = [
    "operating_revenue",
    "gross_profit",
    "operating_profit",
    "total_profit",
    "np_parent_owners",
    "net_profit_cut",
    "net_operate_cashflow",
    "total_assets",
    "se_without_minority",
    "total_shares",
    "basic_eps",
    "eps_weighted",
    "eps_cut_epscut",
    "eps_cut_weighted",
    "roe",
    "roe_weighted",
    "roe_cut",
    "roe_cut_weighted",
    "net_operate_cashflow_per_share",
    "equity_per_share",
    "operating_revenue_yoy",
    "gross_profit_yoy",
    "operating_profit_yoy",
    "total_profit_yoy",
    "np_parent_minority_pany_yoy",
    "ne_t_minority_ty_yoy",
    "net_operate_cash_flow_yoy",
    "total_assets_to_opening",
    "se_without_minority_to_opening",
    "basic_eps_yoy",
    "eps_weighted_yoy",
    "eps_cut_yoy",
    "eps_cut_weighted_yoy",
    "roe_yoy",
    "roe_weighted_yoy",
    "roe_cut_yoy",
    "roe_cut_weighted_yoy",
    "net_operate_cash_flow_per_share_yoy",
    "net_asset_psto_opening",
]

QUARTER_DATE_MAP = {1: (3, 31), 2: (6, 30), 3: (9, 30), 4: (12, 31)}

__TIME_DELTA_MAP__ = {
    "y": relativedelta(years=1),
    "m": relativedelta(months=1),
    "q": relativedelta(months=3),
    "d": relativedelta(days=1),
}


@export_as_api(name="fundamentals")
@export_as_api
class Fundamentals:
    stockcode = FundamentalBase.stockcode
    announce_date = FundamentalBase.announce_date
    income_statement = StkIncomeGen
    balance_sheet = StkBalaGen
    cash_flow = StkCashGen
    cash_flow_statement = StkCashGen
    financial_indicator = AnaStkFinIdx
    eod_derivative_indicator = AnaStkValIdx
    fundamental_base = FundamentalBase
    cash_flow_statement_TTM = CashFlowStatementTTM
    income_statement_TTM = IncomeStatementTTM
    financial_indicator_TTM = FinancialIndicatorTTM


@export_as_api(name="query")
def query_entities(*entities):
    base_list = ["stockcode", "tradedate", "end_date", "announce_date", "rpt_year", "rpt_quarter"]
    columns = [
        Fundamentals.fundamental_base.stockcode,
        Fundamentals.fundamental_base.tradedate,
        Fundamentals.fundamental_base.rpt_year,
        Fundamentals.fundamental_base.rpt_quarter,
    ]
    for ele in entities:
        if isinstance(ele, DeclarativeMeta):
            deprecated_list = deprecated_fundamental_data("data_point")
            query_list = [
                v
                for k, v in ele.__dict__.items()
                if not k.startswith("_") and k not in base_list and k not in deprecated_list
            ]
            columns.extend(query_list)
        elif isinstance(ele, InstrumentedAttribute):
            name = str(ele).split(".")[-1]
            if name in ["stockcode", "tradedate", "rpt_year", "rpt_quarter"]:
                continue
            columns.append(ele)
        else:
            raise ValueError(
                "Invalid metrics to query, it maybe not specify metrics, "
                "please check the metrics in query."
            )

    return Query(columns)


def _compile_query(query):
    comp = query.statement.compile(dialect=mysql.dialect())
    comp_params = comp.params
    params = []
    for k in comp.positiontup:
        v = comp_params[k]
        params.append(escape_item(v, conversions, encoders))

    comp = comp.string
    if "equity_preferred_stock" in comp:
        if "equity_prefer_stock" in comp:
            comp = comp.replace("fundamental_view.equity_prefer_stock,", "")
        comp = comp.replace("equity_preferred_stock", "equity_prefer_stock as equity_preferred_stock")
    elif "equity_prefer_stock" in comp:
        warnings.warn("'equity_prefer_stock' has been deprecated, please use 'equity_preferred_stock'.")

    return comp % tuple(params)


def _unsafe_apply_query_filter(query, trading_dates):
    # TODO this is a hack
    limit, offset = query._limit, query._offset
    query = query.limit(None).offset(None)

    if query._order_by:
        # 对于存在order_by的, 过滤掉 NaN
        def _filter(q, column):
            if isinstance(column, Column):
                return q.filter(column is not None)
            if isinstance(column, UnaryExpression):
                return _filter(q, column.element)
            return q

        for creterion in query._order_by:
            query = _filter(query, creterion)

    query = query.filter(FundamentalBase.tradedate.in_([to_date_int(d) for d in trading_dates]))

    if len(trading_dates) > 1:
        query = query.order_by(FundamentalBase.tradedate.desc())
    return query.limit(limit).offset(offset)


def _parse_arguments(query, quarter, interval):
    if not isinstance(query, Query):
        raise ValueError(
            "The first argument must be a sqlalchemy's Query object. "
            "But what passed in was: " + str(type(query))
        )
    quarters = get_quarters(quarter, interval)
    quarter_dates = [quarter_to_date(y, q) for y, q in quarters]
    query = adjust_query(query, quarter_dates)
    return _compile_query(query), quarters


def quarter_generator(year, quarter):
    while True:
        yield year, quarter
        quarter -= 1
        if quarter == 0:
            quarter = 4
            year -= 1


def year_generator(year, quarter):
    while True:
        yield year, quarter
        year -= 1


def quarter_to_date(year, quarter):
    dates = (331, 630, 930, 1231)
    return year * 10000 + dates[quarter - 1]


def get_quarters(quarter, interval):
    if quarter is None:
        raise ValueError("quarter is required.")

    splited = quarter.lower().split("q")
    if len(splited) != 2:
        raise ValueError("wrong quarter format, use format like '2016q1'")

    year, quarter = int(splited[0]), int(splited[1])
    if not 1 <= quarter <= 4:
        raise ValueError("quarter should be in [1, 4]")

    if interval is not None:
        if not isinstance(interval, str):
            raise ValueError("interval should be a string like 4q, 2y")

        if interval[-1] not in ("y", "q", "Y", "Q"):
            raise ValueError("interval unit should be y(year) or q(quarter)")
        try:
            int(interval[:-1])
        except ValueError:
            raise ValueError("interval should be a string like 4q, 2y")
    if interval is not None:
        generator = (quarter_generator if interval[-1].lower() == "q" else year_generator)(
            year, quarter
        )
        n = int(interval[:-1])
        return list(islice(generator, n))
    else:
        return [(year, quarter)]


def adjust_query(query, quarter_dates):
    # hack
    if str(query._entities[1]).split(".")[-1] == "tradedate":
        query._entities.pop(1)  # remove tradedate entity
    limit, offset = query._limit, query._offset
    query.limit(None).offset(None)

    query = query.filter(FundamentalBase.end_date.in_(quarter_dates))

    if len(quarter_dates) > 1:
        query = query.order_by(FundamentalBase.end_date.desc())
    return query.limit(limit).offset(offset)


def parse_results(records, quarters, expect_df):
    if not records:
        return None

    removed_items_size = 3
    base_fields = ["stockcode", "rpt_year", "rpt_quarter"]

    field_names = base_fields + list(set(records[0].keys()) - set(base_fields))
    items = field_names[removed_items_size:]

    if not expect_df and not is_panel_removed:
        # 只有一个查询日期时, 保持顺序
        stocks = list(set(r[field_names[0]] for r in records))
        stock_index = {s: i for i, s in enumerate(stocks)}
        quarter_index = {q: i for i, q in enumerate(quarters)}

        array = np.ndarray(
            (len(records[0]) - removed_items_size, len(quarters), len(stocks)), dtype=object
        )
        array.fill(np.nan)
        for r in records:
            istock = stock_index[r[field_names[0]]]
            iquarter = quarter_index[(r[field_names[1]], r[field_names[2]])]
            for i in range(removed_items_size, len(r)):
                if field_names[i] == "announce_date":
                    announce_date = r[field_names[i]]
                    array[i - removed_items_size, iquarter, istock] = (
                        np.nan if announce_date is None else announce_date
                    )
                else:
                    array[i - removed_items_size, iquarter, istock] = np.float64(r[field_names[i]])

        s_quarters = ["%dq%d" % (year, quarter) for year, quarter in quarters]
        results = pd.Panel(data=array, items=items, major_axis=s_quarters, minor_axis=stocks)
        item_size, major_size, minor_size = results.shape
        if minor_size == 1:
            ret = results.minor_xs(*stocks)
            return ret[field_names[removed_items_size]] if item_size == 1 else ret
        elif item_size == 1:
            return results[field_names[removed_items_size]]
        elif major_size == 1:
            return results.major_xs(*s_quarters)
        else:
            warnings.warn("Panel is removed after pandas version 0.25.0."
                          " the default value of 'expect_df' will change to True in the future.")
            return results
    else:
        df = pd.DataFrame(records)
        df.rename(columns={"stockcode": "order_book_id"}, inplace=True)
        df["quarter"] = df["rpt_year"].map(str) + "q" + df["rpt_quarter"].map(str)
        df.sort_values(["order_book_id", "quarter"], ascending=[True, False], inplace=True)
        df.set_index(["order_book_id", "quarter"], inplace=True)
        for item in items:
            if item != "announce_date":
                df[item] = df[item].astype(np.float64)
        df = df[items]
        if expect_df:
            return df
        else:
            if len(df.index.get_level_values(0).unique()) == 1:
                df.reset_index(level=0, inplace=True, drop=True)
                df.index.name = None
                df.sort_index(ascending=False, inplace=True)
                if len(df.columns) == 1:
                    df = df[df.columns[0]]
                return df
            elif len(df.columns) == 1:
                field = df.columns[0]
                df = df.unstack(0)[field]
                df.index.name = None
                df.columns.name = None
                df.sort_index(ascending=False, inplace=True)
                return df
            elif len(df.index.get_level_values(1).unique()) == 1:
                df.reset_index(level=1, inplace=True, drop=True)
                df.index.name = None
                return df
            raise_for_no_panel()




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-11-05T18:05:23+0800",
 "dirty": false,
 "error": null,
 "full-revisionid": "0e48433fe5b1f3a64b5db638ed51be290d8a204a",
 "version": "2.9.39.3"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



# -*- coding: utf-8 -*-
import warnings
import datetime

from dateutil.relativedelta import relativedelta
from six import string_types, binary_type, text_type

from rqdatac.utils import listify, to_date, is_panel_removed


def check_quarter(quarter, name=""):
    if not isinstance(quarter, string_types):
        raise ValueError("{}: expect string, got{!r}".format(name, quarter))
    quarter_elm = quarter.lower().split("q")
    if len(quarter_elm) != 2:
        raise ValueError("{}: invalid quarter param, got{!r} please use format like '2016q1'".format(name, quarter))
    if len(quarter_elm[0]) != 4 or len(quarter_elm[1]) != 1:
        raise ValueError("{}: invalid param, got{!r} please use format like '2016q1'".format(name, quarter))
    ensure_int(quarter_elm[1], name)
    ensure_int(quarter_elm[0], name)
    if not (1 <= int(quarter_elm[1]) <= 4 or 1900 <= int(quarter_elm[0]) <= 3000):
        raise ValueError(
            "{}: invalid year/quarter param, got{!r} please use format like '2016q1'".format(name, quarter))


def quarter_string_to_date(quarter_string):
    mappings = {
        "1": "03-31",
        "2": "06-30",
        "3": "09-30",
        "4": "12-31"
    }
    year, quarter = quarter_string.lower().split("q")
    return "{}-{}".format(year, mappings[quarter])


def ensure_list_of_string(s, name=""):
    if isinstance(s, string_types):
        return [s]

    result = list(s)
    for v in result:
        if not isinstance(v, string_types):
            raise ValueError("{}: expect string or list of string, got {!r}".format(name, v))
    return result


def ensure_list(s, name="", except_type=None):
    result = list(s)
    if except_type is None:
        return result
    for v in result:
        if not isinstance(v, except_type):
            raise ValueError("{}: expect {!r}, got {!r}".format(name, except_type, v))
    return result


def ensure_string(s, name="", decoding="utf-8"):
    if isinstance(s, binary_type):
        return s.decode(decoding)
    if not isinstance(s, text_type):
        raise ValueError("{}: expect a string, got {!r}".format(name, s))
    return s


def ensure_string_in(s, should_in, name="", decoding="utf-8"):
    s = ensure_string(s, name, decoding)
    if s not in should_in:
        raise TypeError("{}: expect value in {!r}".format(name, should_in))
    return s


def check_type(s, t, name=""):
    if not isinstance(s, t):
        raise ValueError("{}: expect value in type {}, got {!r}.".format(name, t, s))


def ensure_int(s, name=""):
    try:
        return int(s)
    except TypeError:
        raise ValueError("{}: expect int value, got {!r}.".format(name, s))


def check_items_in_container(items, should_in, name):
    items = listify(items)
    for item in items:
        if item not in should_in:
            raise ValueError(
                "{}: got invalided value {}, choose any in {}".format(name, item, should_in)
            )


def ensure_order(items, ordered):
    # type: (list, iter) -> list
    items = set(items)
    return [i for i in ordered if i in items]


def ensure_date_str(date):
    # type: (...) -> str
    date = to_date(date)
    return "%04d-%02d-%02d" % (date.year, date.month, date.day)


def ensure_date_int(date):
    date = to_date(date)
    return date.year * 10000 + date.month * 100 + date.day


def ensure_date_or_today_int(date):
    if date:
        return ensure_date_int(date)
    return _to_date_int(datetime.datetime.today())


def _to_date_int(date):
    # type: (datetime.datetime or datetime.date) -> int
    return date.year * 10000 + date.month * 100 + date.day


def ensure_date_range(start_date, end_date, delta=relativedelta(months=3)):
    if start_date is None and end_date is None:
        return _to_date_int(datetime.date.today() - delta), _to_date_int(datetime.date.today())

    if start_date is None:
        end_date = to_date(end_date)
        return _to_date_int(end_date - delta), _to_date_int(end_date)

    if end_date is None:
        start_date = to_date(start_date)
        return _to_date_int(start_date), _to_date_int(start_date + delta)

    s, e = ensure_date_int(start_date), ensure_date_int(end_date)
    if s > e:
        raise ValueError("invalid date range: [{!r}, {!r}]".format(start_date, end_date))
    return s, e


def ensure_dates_base_on_listed_date(instrument, start_date, end_date, market):
    from rqdatac.services.calendar import get_previous_trading_date, get_latest_trading_date
    if to_date(instrument.listed_date) > datetime.date.today():
        raise ValueError("instrument {} is not listed yet".format(instrument.order_book_id))

    if start_date is None:
        start_date = instrument.listed_date
    elif to_date(start_date) < to_date(instrument.listed_date):
        start_date = to_date(instrument.listed_date)
    elif instrument.de_listed_date != "0000-00-00" and to_date(start_date) >= to_date(instrument.de_listed_date):
        warnings.warn("{} has been delisted on {}".format(instrument.order_book_id, instrument.de_listed_date))
        return None, None

    if end_date is None:
        if instrument.de_listed_date != "0000-00-00":
            end_date = get_previous_trading_date(instrument.de_listed_date, market=market)
        else:
            end_date = get_latest_trading_date(market=market)
    elif instrument.de_listed_date != "0000-00-00" and to_date(end_date) >= to_date(instrument.de_listed_date):
        warnings.warn("{} has been delisted on {}".format(instrument.order_book_id, instrument.de_listed_date))
        end_date = get_previous_trading_date(instrument.de_listed_date, market=market)
    return start_date, end_date


def ensure_instruments(order_book_ids, type=None, market="cn"):
    order_book_ids = ensure_list_of_string(order_book_ids)
    from rqdatac.services.basic import _all_instruments_dict

    all_instruments = _all_instruments_dict(market)
    result = []
    for ob in set(order_book_ids):
        i = all_instruments.get(ob)
        if not i:
            warnings.warn("invalid order_book_id: {}".format(ob), stacklevel=0)
            continue
        if type is not None and i.type != type:
            warnings.warn(
                "expect {} instrument, got {}({}), ignored".format(type, i.type, ob), stacklevel=0
            )
            continue
        result.append(i)
    if not result:
        raise ValueError("order_book_ids: at least one valid instrument expected, got none")
    return result


def ensure_instruments_in_order(order_book_ids, type=None, market="cn"):
    order_book_ids = ensure_list_of_string(order_book_ids)
    from rqdatac.services.basic import _all_instruments_dict

    all_instruments = _all_instruments_dict(market)
    result = []
    unique = set()
    result_append = result.append
    unique_add = unique.add
    for ob in order_book_ids:
        i = all_instruments.get(ob)
        if not i:
            warnings.warn("invalid order_book_id: {}".format(ob), stacklevel=0)
            continue
        if type is not None and i.type != type:
            warnings.warn(
                "expect {} instrument, got {}({}), ignored".format(type, i.type, ob), stacklevel=0
            )
            continue
        if ob in unique:
            warnings.warn("duplicated order_book_id: {}".format(ob), stacklevel=0)
            continue
        unique_add(ob)
        result_append(i)
    if not result:
        raise ValueError("order_book_ids: at least one valid instrument expected, got none")
    return result


def ensure_order_book_ids(order_book_ids, type=None, market="cn"):
    order_book_ids = ensure_list_of_string(order_book_ids)
    from rqdatac.services.basic import _all_instruments_dict

    all_instruments = _all_instruments_dict(market)
    result = []
    for ob in set(order_book_ids):
        i = all_instruments.get(ob)
        if not i:
            warnings.warn("invalid order_book_id: {}".format(ob))
            continue
        if type is not None and i.type != type:
            warnings.warn("expect {} instrument, got {}({}), ignored".format(type, i.type, ob))
            continue
        result.append(i.order_book_id)
    if not result:
        raise ValueError("order_book_ids: at least one valid instrument expected, got none")
    return result


def ensure_order_book_id(ob, type=None, market="cn"):
    ob = ensure_string(ob, "order_book_id")
    from rqdatac.services.basic import _all_instruments_dict

    all_instruments = _all_instruments_dict(market)
    try:
        i = all_instruments[ob]
    except KeyError as e:
        raise ValueError("invalid order_book_id: {!s}".format(e))
    if type is not None and i.type != type:
        raise ValueError("expect {} instrument, got {}({}), ignored".format(type, i.type, ob))
    return i.order_book_id


def ensure_trading_date(date):
    from rqdatac.services.calendar import get_trading_dates, get_previous_trading_date, get_next_trading_date

    trading_dates = get_trading_dates(get_previous_trading_date(date), get_next_trading_date(date))
    if date not in trading_dates:
        raise ValueError(
            "expect a trading date, got {}, for reference {}".format(date.strftime("%Y%m%d"), trading_dates)
        )
    return date


def raise_for_no_panel(expect_df=False):
    if not expect_df and is_panel_removed:
        raise RuntimeError("Panel has been removed since pandas's version >= 0.25, "
                           "you could set 'expect_df=True' and then get a MultiIndex DataFrame")


def ensure_expect_df(expect_df):
    if not expect_df and is_panel_removed:
        raise RuntimeError("Panel has been removed since pandas's version >= 0.25, "
                           "you could set 'expect_df=True' and then get a MultiIndex DataFrame")



import rqdatac.services.financial
import rqdatac.services.market_data
import rqdatac.services.live_md_client
import rqdatac.services.basic
import rqdatac.services.get_capital_flow
import rqdatac.services.constant
import rqdatac.services.calendar
import rqdatac.services.factor
import rqdatac.services.concept
import rqdatac.services.xueqiu
import rqdatac.services.index
import rqdatac.services.ksh_auction_info
import rqdatac.services.future
import rqdatac.services.structured_fund
import rqdatac.services.convertible
import rqdatac.services.tmall
import rqdatac.services.options
import rqdatac.services.shenwan
import rqdatac.services.stock_status
import rqdatac.services.live
import rqdatac.client
import rqdatac.services.get_price

__version__: str = ...
init = rqdatac.client.init
reset = rqdatac.client.reset
initialized = rqdatac.client.initialized
concept_list = rqdatac.services.concept.concept_list
concept = rqdatac.services.concept.concept
concept_names = rqdatac.services.concept.concept_names
shenwan_industry = rqdatac.services.shenwan.shenwan_industry
shenwan_instrument_industry = rqdatac.services.shenwan.shenwan_instrument_industry
zx_industry = rqdatac.services.shenwan.zx_industry
zx_instrument_industry = rqdatac.services.shenwan.zx_instrument_industry
get_industry = rqdatac.services.shenwan.get_industry
get_instrument_industry = rqdatac.services.shenwan.get_instrument_industry
get_industry_mapping = rqdatac.services.shenwan.get_industry_mapping
industry_code = rqdatac.services.constant.IndustryCode
IndustryCode = rqdatac.services.constant.IndustryCode
sector_code = rqdatac.services.constant.SectorCode
SectorCode = rqdatac.services.constant.SectorCode
get_trading_dates = rqdatac.services.calendar.get_trading_dates
get_next_trading_date = rqdatac.services.calendar.get_next_trading_date
get_previous_trading_date = rqdatac.services.calendar.get_previous_trading_date
get_latest_trading_date = rqdatac.services.calendar.get_latest_trading_date
trading_date_offset = rqdatac.services.calendar.trading_date_offset
is_trading_date = rqdatac.services.calendar.is_trading_date
has_night_trading = rqdatac.services.calendar.has_night_trading
id_convert = rqdatac.services.basic.id_convert
instruments = rqdatac.services.basic.instruments
all_instruments = rqdatac.services.basic.all_instruments
sector = rqdatac.services.basic.sector
industry = rqdatac.services.basic.industry
get_future_contracts = rqdatac.services.basic.get_future_contracts


class futures:
    get_commission_margin = rqdatac.services.future.get_commission_margin
    get_contracts = rqdatac.services.basic.get_contracts
    get_dominant = rqdatac.services.future.get_dominant
    get_dominant_price = rqdatac.services.future.get_dominant_price
    get_member_rank = rqdatac.services.future.get_member_rank
    get_warehouse_stocks = rqdatac.services.future.get_warehouse_stocks
    get_contract_multiplier = rqdatac.services.future.get_contract_multiplier
    get_ex_factor = rqdatac.services.future.get_ex_factor


jy_instrument_industry = rqdatac.services.basic.jy_instrument_industry


class econ:
    get_factors = rqdatac.services.basic.get_factors
    get_money_supply = rqdatac.services.basic.get_money_supply
    get_reserve_ratio = rqdatac.services.basic.get_reserve_ratio


get_main_shareholder = rqdatac.services.basic.get_main_shareholder
get_current_news = rqdatac.services.basic.get_current_news
get_trading_hours = rqdatac.services.basic.get_trading_hours
get_private_placement = rqdatac.services.basic.get_private_placement
get_share_transformation = rqdatac.services.basic.get_share_transformation


class user:
    get_quota = rqdatac.services.basic.get_quota


get_update_status = rqdatac.services.basic.get_update_status
info = rqdatac.services.basic.info
get_basic_info = rqdatac.services.basic.get_basic_info


class convertible:
    all_instruments = rqdatac.services.convertible.all_instruments
    get_call_info = rqdatac.services.convertible.get_call_info
    get_cash_flow = rqdatac.services.convertible.get_cash_flow
    get_conversion_info = rqdatac.services.convertible.get_conversion_info
    get_conversion_price = rqdatac.services.convertible.get_conversion_price
    get_credit_rating = rqdatac.services.convertible.get_credit_rating
    get_indicators = rqdatac.services.convertible.get_indicators
    get_industry = rqdatac.services.convertible.get_industry
    get_instrument_industry = rqdatac.services.convertible.get_instrument_industry
    get_latest_rating = rqdatac.services.convertible.get_latest_rating
    get_put_info = rqdatac.services.convertible.get_put_info
    instruments = rqdatac.services.convertible.instruments
    is_suspended = rqdatac.services.convertible.is_suspended
    rating = rqdatac.services.convertible.rating


get_dominant_future = rqdatac.services.future.get_dominant_future
future_commission_margin = rqdatac.services.future.future_commission_margin
get_future_member_rank = rqdatac.services.future.get_future_member_rank
current_stock_connect_quota = rqdatac.services.stock_status.current_stock_connect_quota
get_stock_connect_quota = rqdatac.services.stock_status.get_stock_connect_quota
is_st_stock = rqdatac.services.stock_status.is_st_stock
_is_st_stock = rqdatac.services.stock_status._is_st_stock
is_suspended = rqdatac.services.stock_status.is_suspended
get_stock_connect = rqdatac.services.stock_status.get_stock_connect
get_securities_margin = rqdatac.services.stock_status.get_securities_margin
get_margin_stocks = rqdatac.services.stock_status.get_margin_stocks
get_shares = rqdatac.services.stock_status.get_shares
get_allotment = rqdatac.services.stock_status.get_allotment
current_snapshot = rqdatac.services.live.current_snapshot
get_ticks = rqdatac.services.live.get_ticks
current_minute = rqdatac.services.live.current_minute
get_live_ticks = rqdatac.services.live.get_live_ticks
get_price = rqdatac.services.get_price.get_price
LiveMarketDataClient = rqdatac.services.live_md_client.LiveMarketDataClient
get_all_factor_names = rqdatac.services.factor.get_all_factor_names
get_factor = rqdatac.services.factor.get_factor
get_factor_return = rqdatac.services.factor.get_factor_return
get_factor_exposure = rqdatac.services.factor.get_factor_exposure
get_style_factor_exposure = rqdatac.services.factor.get_style_factor_exposure
get_descriptor_exposure = rqdatac.services.factor.get_descriptor_exposure
get_stock_beta = rqdatac.services.factor.get_stock_beta
get_factor_covariance = rqdatac.services.factor.get_factor_covariance
get_specific_return = rqdatac.services.factor.get_specific_return
get_specific_risk = rqdatac.services.factor.get_specific_risk
get_index_factor_exposure = rqdatac.services.factor.get_index_factor_exposure
Financials = rqdatac.services.financial.Financials
financials = rqdatac.services.financial.Financials
get_financials = rqdatac.services.financial.get_financials
PitFinancials = rqdatac.services.financial.PitFinancials
pit_financials = rqdatac.services.financial.PitFinancials
get_pit_financials = rqdatac.services.financial.get_pit_financials
get_pit_financials_ex = rqdatac.services.financial.get_pit_financials_ex
get_fundamentals = rqdatac.services.financial.get_fundamentals
deprecated_fundamental_data = rqdatac.services.financial.deprecated_fundamental_data
current_performance = rqdatac.services.financial.current_performance
performance_forecast = rqdatac.services.financial.performance_forecast
Fundamentals = rqdatac.services.financial.Fundamentals
fundamentals = rqdatac.services.financial.Fundamentals
query = rqdatac.services.financial.query_entities
get_capital_flow = rqdatac.services.get_capital_flow.get_capital_flow
get_open_auction_info = rqdatac.services.get_capital_flow.get_open_auction_info
index_components = rqdatac.services.index.index_components
index_weights = rqdatac.services.index.index_weights
index_indicator = rqdatac.services.index.index_indicator
get_ksh_auction_info = rqdatac.services.ksh_auction_info.get_ksh_auction_info
get_split = rqdatac.services.market_data.get_split
get_dividend = rqdatac.services.market_data.get_dividend
get_dividend_info = rqdatac.services.market_data.get_dividend_info
get_ex_factor = rqdatac.services.market_data.get_ex_factor
get_turnover_rate = rqdatac.services.market_data.get_turnover_rate
get_price_change_rate = rqdatac.services.market_data.get_price_change_rate
get_yield_curve = rqdatac.services.market_data.get_yield_curve
get_block_trade = rqdatac.services.market_data.get_block_trade
get_exchange_rate = rqdatac.services.market_data.get_exchange_rate
get_temporary_code = rqdatac.services.market_data.get_temporary_code


class options:
    get_contract_property = rqdatac.services.options.get_contract_property
    get_contracts = rqdatac.services.options.get_contracts
    get_greeks = rqdatac.services.options.get_greeks




class fenji:
    get = rqdatac.services.structured_fund.get
    get_a_by_interest_rule = rqdatac.services.structured_fund.get_a_by_interest_rule
    get_a_by_yield = rqdatac.services.structured_fund.get_a_by_yield
    get_all = rqdatac.services.structured_fund.get_all


ecommerce = rqdatac.services.tmall.ecommerce


class xueqiu:
    history = rqdatac.services.xueqiu.history
    top_stocks = rqdatac.services.xueqiu.top_stocks


get_consensus_indicator = rqdatac.services.consensus.get_consensus_indicator
get_consensus_price = rqdatac.services.consensus.get_consensus_price



# -*- coding: utf-8 -*-
import datetime
import warnings
from itertools import chain
from collections import OrderedDict

import pandas as pd
import numpy as np

from rqdatac.services.calendar import (
    get_previous_trading_date,
    get_next_trading_date,
    get_trading_dates,
)
from rqdatac.validators import (
    ensure_list_of_string,
    ensure_date_int,
    ensure_date_range,
    ensure_string,
    ensure_string_in,
    check_items_in_container,
    ensure_order_book_ids,
    ensure_order_book_id,
)
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api


@export_as_api
def get_all_factor_names(market="cn"):
    """获取因子列表

    :return:
        list

    :param market:  (Default value = "cn")

    """
    return get_client().execute("get_all_factor_names", market)


@export_as_api
def get_factor(order_book_ids, factor, start_date=None, end_date=None, universe=None, expect_df=True, **kwargs):
    """获取因子

    :param order_book_ids: 股票代码或代码列表
    :param factor: 如 'total_income'
    :param date: 如 date='2015-01-05', 默认为前一交易日
    :param start_date: 开始日期'2015-01-05', 默认为前一交易日, 最小起始日期为'2000-01-04'
    :param end_date: 结束日期
    :param universe: 股票池，默认为全A股
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :returns: pd.DataFrame
    """

    order_book_ids = ensure_order_book_ids(order_book_ids, type="CS")
    order_book_ids = list(set(order_book_ids))

    factor = ensure_list_of_string(factor)
    factor = list(OrderedDict.fromkeys(factor))

    if start_date and end_date:
        start_date, end_date = ensure_date_range(start_date, end_date, datetime.timedelta(days=15))
        if start_date < 20000104:
            warnings.warn("start_date is earlier than 2000-01-04, adjusted to 2000-01-04")
            start_date = 20000104
    elif start_date:
        raise ValueError("Expect end_date")
    elif end_date:
        raise ValueError("Expect start_date")
    else:
        date = kwargs.pop("date", None)
        date = ensure_date_int(date or get_previous_trading_date(datetime.date.today()))
        start_date = end_date = date

    if kwargs:
        raise ValueError('unknown kwargs: {}'.format(kwargs))

    if universe is not None:
        universe = ensure_string(universe, "universe")
        if universe != "all":
            universe = ensure_order_book_id(universe, type="INDX")
            comp_data = get_client().execute('index_components_v2', universe, start_date, end_date, market='cn')
            allowed_order_book_ids = set()
            for comp in comp_data:
                allowed_order_book_ids.update(comp['component_ids'])
            not_permit_order_book_ids = set(order_book_ids) - allowed_order_book_ids
            if not_permit_order_book_ids:
                warnings.warn(
                    "%s not in universe pool from %s to %s, value of those order_book_ids may be NaN"
                    % (not_permit_order_book_ids, start_date, end_date)
                )

    data = get_client().execute(
        "get_factor_from_store", order_book_ids, factor, start_date, end_date, universe=universe
    )

    if not data:
        return

    factor_value_length = len(data[0][2])
    if factor_value_length == 0:
        return

    dates = pd.to_datetime(get_trading_dates(start_date, end_date))
    days = len(dates)
    if days > factor_value_length:
        _get_factor_warning_msg(dates[factor_value_length], dates[-1])
        dates = dates[0:factor_value_length]

    if expect_df or len(factor) > 1:
        order_book_id_index_map = {o: i for i, o in enumerate(order_book_ids)}
        factor_index_map = {f: i for i, f in enumerate(factor)}
        arr = np.full((len(order_book_ids) * days, len(factor)), np.nan)

        for order_book_id, factor_name, values in data:
            if not values:
                continue
            value_length = min(days, len(values))
            order_book_id_index = order_book_id_index_map[order_book_id]
            factor_index = factor_index_map[factor_name]
            start = order_book_id_index * days
            arr[start: start + value_length, factor_index] = values[-value_length:]

        multi_index = pd.MultiIndex.from_product([order_book_ids, dates], names=["order_book_id", "date"])
        df = pd.DataFrame(index=multi_index, columns=factor, data=arr)
        return df

    order_book_id_index_map = {o: i for i, o in enumerate(order_book_ids)}
    arr = np.full((days, len(order_book_ids)), np.nan)
    for order_book_id, _, values in data:
        arr[:len(values), order_book_id_index_map[order_book_id]] = values
    df = pd.DataFrame(index=dates, columns=order_book_ids, data=arr)

    if len(df.index) == 1:
        return df.iloc[0]
    if len(df.columns) == 1:
        return df[df.columns[0]]
    return df


def _get_factor_warning_msg(start_date, end_date):
    if start_date == end_date:
        end_date = end_date.strftime("%Y%m%d")
        warnings.warn("{} calculation not completed".format(end_date))
    else:
        start_date = start_date.strftime("%Y%m%d")
        end_date = end_date.strftime("%Y%m%d")
        warnings.warn(
            "{} - {} calculation not completed".format(start_date, end_date))


_UNIVERSE_MAPPING = {
    "whole_market": "whole_market",
    "000300.XSHG": "csi_300",
    "000905.XSHG": "csi_500",
    "000906.XSHG": "csi_800",
}

_METHOD_MAPPING = {"explicit": "explicit_factor_return", "implicit": "implicit_factor_return"}


@export_as_api
def get_factor_return(
    start_date, end_date, factors=None, universe="whole_market", method="implicit", industry_mapping=True, market="cn",
):
    """获取因子收益率数据

    :param start_date: 开始日期（例如：‘2017-03-03’)
    :param end_date: 结束日期（例如：‘2017-03-20’)
    :param factors: 因子。默认获取全部因子的因子收益率
        当 method 参数取值为'implicit' ，可返回全部因子（风格、行业、市场联动）的隐式因子收益率；
        当 method 参数取值为'explicit' , 只返回风格因子的显式因子收益率。具体因子名称见说明文档 (Default value = None)
    :param universe: 股票池。默认调用全市场收益率。可选沪深300（‘000300.XSHG’）、中证500（'000905.XSHG'）
        、以及中证800（'000906.XSHG'） (Default value = "whole_market")
    :param method: 计算方法。默认为'implicit'（隐式因子收益率），可选'explicit'（显式风格因子收益率) (Default value = "implicit")
    :param market: 地区代码， 现在仅支持 'cn' (Default value = "cn")
    :param industry_mapping(bool): 是否按 2014 年后的申万行业分类标 准计算行业收益率.默认为 True.
        若取值为 False,则 2014 年前的行业 收益率按旧行业分类标准计算
    :returns: pd.DataFrame. index 为日期，column 为因子字段名称。

    Usage example::
        # 获取介于2017-03-03 到 2017-03-20到隐式因子收益率数据
        get_factor_return('2017-03-03', '2017-03-20')

    """
    start_date, end_date = ensure_date_range(start_date, end_date)

    if factors:
        factors = ensure_list_of_string(factors)

    method = ensure_string(method)
    if method not in _METHOD_MAPPING:
        raise ValueError("invalid method: {!r}, valid: explicit, implicit".format(method))
    method = _METHOD_MAPPING[method]

    if universe not in _UNIVERSE_MAPPING:
        raise ValueError(
            "invalid universe: {!r}, valid: {}".format(universe, list(_UNIVERSE_MAPPING.keys()))
        )
    universe = _UNIVERSE_MAPPING[universe]

    df = get_client().execute(
        "get_factor_return", start_date, end_date, factors, universe, method, market=market, industry_mapping=industry_mapping
    )
    if not df:
        return None
    df = pd.DataFrame(df)
    # convert to required format.
    df = df.pivot(index="date", columns="factor")[universe]
    df.sort_index(inplace=True)
    return df


_EXPOSURE_FACTORS = (
    "beta",
    "growth",
    "liquidity",
    "leverage",
    "momentum",
    "reversal",
    "size",
    "yield",
    "volatility",
    "value",
)


@export_as_api
def get_factor_exposure(order_book_ids, start_date=None, end_date=None, factors=None, industry_mapping=True, market="cn"):
    """获取因子暴露度

    :param order_book_ids: 股票代码或代码列表
    :param start_date: 如'2013-01-04' (Default value = None)
    :param end_date: 如'2014-01-04' (Default value = None)
    :param factors: 如'yield', 'beta', 'volatility' (Default value = None)
    :param market: 地区代码, 如'cn' (Default value = "cn")
    :param industry_mapping (bool): 是否按 2014 年后的申万行业分类标 准计算行业暴露度.默认为 True.
        若取值为 False,则 2014 年前的行业 暴露度按旧行业分类标准计算
    :returns: MultiIndex DataFrame. index 第一个 level 为 order_book_id，第 二个 level 为 date，columns 为因子字段名称
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if not order_book_ids:
        raise ValueError("no valid order_book_id found")

    start_date, end_date = ensure_date_range(start_date, end_date)

    if factors is not None:
        factors = ensure_list_of_string(factors)
        check_items_in_container(factors, exposure_factors, "factors")
    results = get_client().execute(
        "get_factor_exposure", order_book_ids, start_date, end_date, factors, market,
        industry_mapping
    )

    if not results:
        return None
    index_pairs = []
    data = []

    fields = [
        field for field in results[0].keys() if field not in ("order_book_id", "date", "industry")
    ]

    industry_factors = _get_all_industries(start_date, end_date, industry_mapping)
    for result in results:
        index_pairs.append((result["date"], result["order_book_id"]))
        row_data = [result.get(field, np.nan) for field in fields]

        # 填充行业因子数据
        for industry in industry_factors:
            if result["industry"] == industry:
                industry_label = 1
            else:
                industry_label = 0
            row_data.append(industry_label)

        data.append(row_data)

    index = pd.MultiIndex.from_tuples(index_pairs, names=["date", "order_book_id"])
    fields.extend(industry_factors)
    result_df = pd.DataFrame(columns=fields, index=index, data=data)
    result_df.sort_index(level=1, inplace=True)

    no_data_book_id = set(order_book_ids) - set(result_df.index.levels[1])
    if no_data_book_id:
        warnings.warn("No data for this order_book_id :{}".format(no_data_book_id))

    if factors is not None:
        exists_factors = list(set(factors) & set(result_df))
        # shenwan industry has re-structured on 2014-01-01, so if industry_mapping is true
        # and given factor is `old shenwan industry`, the `old shenwan industry` will not exists
        # in the result dataframe columns.
        if not exists_factors:
            return None
        return result_df[exists_factors]
    return result_df


def _get_all_industries(start_date, end_date, industry_mapping=True):
    """ 获取在指定时间区间内申万的所有行业分类

    :param start_date (int): 开始日期
    :param end_date (int): 结束日期
    :param industry_mapping (bool): 是否要使用行业映射。如果使用行业映射的话, 那么2014年
        以前的行业分类将会全部都用2014年以后的行业分类代替
    """
    if industry_mapping is True or start_date >= 20140101:
        return SHENWAN_INDUSTRY_NAME_AFTER_2014
    # 申万在2014年1月1号进行过一次行业重组
    if end_date <= 20140101:
        return SHENWAN_INDUSTRY_NAME_BEFORE_2014
    else:
        # 时间区间会跨过 2014-01-01
        return list(
            set(chain(SHENWAN_INDUSTRY_NAME_BEFORE_2014, SHENWAN_INDUSTRY_NAME_AFTER_2014))
        )


exposure_factors = [
    "residual_volatility",
    "growth",
    "liquidity",
    "beta",
    "non_linear_size",
    "leverage",
    "earnings_yield",
    "size",
    "momentum",
    "book_to_price",
    "comovement",
]

SHENWAN_INDUSTRY_NAME_BEFORE_2014 = [
    u"金融服务",
    u"房地产",
    u"医药生物",
    u"有色金属",
    u"餐饮旅游",
    u"综合",
    u"建筑建材",
    u"家用电器",
    u"交运设备",
    u"食品饮料",
    u"电子",
    u"信息设备",
    u"交通运输",
    u"轻工制造",
    u"公用事业",
    u"机械设备",
    u"纺织服装",
    u"农林牧渔",
    u"商业贸易",
    u"化工",
    u"信息服务",
    u"采掘",
    u"黑色金属",
]

SHENWAN_INDUSTRY_NAME_AFTER_2014 = [
    u"农林牧渔",
    u"采掘",
    u"化工",
    u"钢铁",
    u"有色金属",
    u"电子",
    u"家用电器",
    u"食品饮料",
    u"纺织服装",
    u"轻工制造",
    u"医药生物",
    u"公用事业",
    u"交通运输",
    u"房地产",
    u"商业贸易",
    u"休闲服务",
    u"综合",
    u"建筑材料",
    u"建筑装饰",
    u"电气设备",
    u"国防军工",
    u"计算机",
    u"传媒",
    u"通信",
    u"银行",
    u"非银金融",
    u"汽车",
    u"机械设备",
]

exposure_factors.extend(SHENWAN_INDUSTRY_NAME_BEFORE_2014)
exposure_factors.extend(SHENWAN_INDUSTRY_NAME_AFTER_2014)

_STYLE_FACTORS = {
    "residual_volatility",
    "growth",
    "liquidity",
    "beta",
    "non_linear_size",
    "leverage",
    "earnings_yield",
    "size",
    "momentum",
    "book_to_price"
}


@export_as_api
def get_style_factor_exposure(order_book_ids, start_date, end_date, factors=None, market="cn"):
    """获取个股风格因子暴露度

    :param order_book_ids: 证券代码（例如：‘600705.XSHG’）
    :param start_date: 开始日期（例如：‘2017-03-03’）
    :param end_date: 结束日期（例如：‘2017-03-20’）
    :param factors: 风格因子。默认调用全部因子的暴露度（'all'）。
        具体因子名称见说明文档 (Default value = None)
    :param market:  (Default value = "cn")

    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)
    if factors is not None:
        factors = ensure_list_of_string(factors)
        check_items_in_container(factors, _STYLE_FACTORS, "factors")

    df = get_client().execute(
        "get_style_factor_exposure", order_book_ids, start_date, end_date, factors, market=market
    )
    if not df:
        return
    return pd.DataFrame(df).set_index(["order_book_id", "date"]).sort_index(level=1)


_DESCRIPTORS = {
    "daily_standard_deviation",
    "cumulative_range",
    "historical_sigma",
    "one_month_share_turnover",
    "three_months_share_turnover",
    "twelve_months_share_turnover",
    "earnings_to_price_ratio",
    "cash_earnings_to_price_ratio",
    "market_leverage",
    "debt_to_assets",
    "book_leverage",
    "sales_growth",
    "earnings_growth",
    "predicted_earning_to_price",
    "short_term_predicted_earnings_growth",
    "long_term_predicted_earnings_growth"
}


@export_as_api
def get_descriptor_exposure(order_book_ids, start_date, end_date, descriptors=None, market="cn"):
    """获取个股细分因子暴露度

    :param order_book_ids: 证券代码（例如：‘600705.XSHG’）
    :param start_date: 开始日期（例如：‘2017-03-03’）
    :param end_date: 结束日期（例如：‘2017-03-20’）
    :param descriptors: 细分风格因子。默认调用全部因子的暴露度（'all'）。
        具体细分因子名称见说明文档 (Default value = None)
    :param market:  (Default value = "cn")
    :returns: MultiIndex DataFrame. index 第一个 level 为 order_book_id，第 二个 level 为 date，column 为细分风格因子字段名称。
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)
    if descriptors is not None:
        if descriptors == "all":
            descriptors = None
        else:
            descriptors = ensure_list_of_string(descriptors)
            check_items_in_container(descriptors, _DESCRIPTORS, "descriptors")

    df = get_client().execute(
        "get_descriptor_exposure", order_book_ids, start_date, end_date, descriptors, market=market
    )
    if not df:
        return
    return pd.DataFrame(df).set_index(["order_book_id", "date"]).sort_index(level=1)


@export_as_api
def get_stock_beta(order_book_ids, start_date, end_date, benchmark="000300.XSHG", market="cn"):
    """获取个股相对于基准的贝塔

    :param order_book_ids: 证券代码（例如：‘600705.XSHG’）
    :param start_date: 开始日期（例如：‘2017-03-03’)
    :param end_date: 结束日期（例如：‘2017-03-20’）
    :param benchmark: 基准指数。默认为沪深300（‘000300.XSHG’）
        可选上证50（'000016.XSHG'）、中证500（'000905.XSHG'）、
        中证800（'000906.XSHG'）以及中证全指（'000985.XSHG'） (Default value = "000300.XSHG")
    :param market:  (Default value = "cn")
    :returns: pandas.DataFrame，index 为日期，column 为个股的 order_book_id
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)

    all_benchmark = ("000300.XSHG", "000016.XSHG", "000905.XSHG", "000906.XSHG", "000985.XSHG")
    benchmark = ensure_string(benchmark, "benchmark")
    check_items_in_container(benchmark, all_benchmark, "benchmark")
    benchmark = benchmark.replace(".", "_")
    df = get_client().execute(
        "get_stock_beta", order_book_ids, start_date, end_date, benchmark, market=market
    )
    if not df:
        return
    df = pd.DataFrame(df)
    df = df.pivot(index="date", columns="order_book_id", values=benchmark).sort_index()
    return df


def get_eigenfactor_adjusted_covariance(date, horizon='daily'):
    """ 获取因子协方差矩阵（特征因子调整）

    :param date: str 日期（例如：‘2017-03-20’）
    :param horizon: str 预测期限。默认为日度（'daily'），可选月度（‘monthly’）或季度（'quarterly'）。

    :return: pandas.DataFrame，其中 index 和 column 均为因子名称。
    """
    date = get_previous_trading_date(get_next_trading_date(date))
    date = ensure_date_int(date)
    ensure_string_in(horizon, HORIZON_CONTAINER, 'horizon')

    df = get_client().execute('get_eigenfactor_adjusted_covariance', date, horizon)
    if not df:
        return
    df = pd.DataFrame(df)
    df.drop("date", axis=1, inplace=True)
    return df.reindex(columns=df.index)


@export_as_api
def get_factor_covariance(date, horizon='daily'):
    """ 获取因子协方差矩阵

    :param date: str 日期（例如：‘2017-03-20’）
    :param horizon: str 预测期限。默认为日度（'daily'），可选月度（‘monthly’）或季度（'quarterly'）。

    :return: pandas.DataFrame，其中 index 和 column 均为因子名称。
    """
    date = get_previous_trading_date(get_next_trading_date(date))
    date = ensure_date_int(date)
    ensure_string_in(horizon, HORIZON_CONTAINER, 'horizon')

    df = get_client().execute('get_factor_covariance', date, horizon)
    if not df:
        return
    df = pd.DataFrame(df)
    df.drop("date", axis=1, inplace=True)
    return df.reindex(columns=df.index)


@export_as_api
def get_specific_return(order_book_ids, start_date, end_date):
    """ 获取个股特异收益率

    :param order_book_ids	str or [list of str]	证券代码（例如：‘600705.XSHG’）
    :param start_date	    str                 	开始日期（例如：‘2017-03-03’）
    :param end_date	        str	                    结束日期（例如：‘2017-03-20’）

    :return: pandas.DataFrame，其中 index 为date, column 为 order_book_ids。
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)

    df = get_client().execute('get_specific_return', order_book_ids, start_date, end_date)
    if not df:
        return
    df = pd.DataFrame(df)
    df = df.pivot(index='date', columns='order_book_id', values="specific_return").sort_index()
    return df


@export_as_api
def get_specific_risk(order_book_ids, start_date, end_date, horizon='daily'):
    """ 获取个股特异方差

    :param order_book_ids	str or [list of str]	证券代码（例如：‘600705.XSHG’）
    :param start_date	    str                 	开始日期（例如：‘2017-03-03’）
    :param end_date	        str	                    结束日期（例如：‘2017-03-20’）
    :param horizon	        str	    预测期限。默认为日度（'daily'），可选月度（‘monthly’）或季度（'quarterly'）

    :return: pandas.DataFrame，其中 index 为date, column 为 order_book_ids。
    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)
    ensure_string_in(horizon, HORIZON_CONTAINER, 'horizon')

    df = get_client().execute('get_specific_risk', order_book_ids, start_date, end_date, horizon)
    if not df:
        return
    df = pd.DataFrame(df)
    df = df.pivot(index="date", columns="order_book_id", values="specific_risk").sort_index()
    return df


def get_cross_sectional_bias(start_date, end_date, type='factor'):
    """ 获取横截面偏差系数

    :param order_book_ids	str or [list of str]	证券代码（例如：‘600705.XSHG’）
    :param start_date	    str                 	开始日期（例如：‘2017-03-03’）
    :param end_date	        str	                    结束日期（例如：‘2017-03-20’）
    :param type	            str	                    默认为 'factor'，可选 'specific'

    :return: pandas.DataFrame，其中 index 为date, column 包含 'daily'、'monthly'  和 'quarterly' 三个字段。
    """
    start_date, end_date = ensure_date_range(start_date, end_date)
    ensure_string_in(type, ['factor', 'specific'], 'horizon')

    df = get_client().execute('get_cross_sectional_bias', start_date, end_date, type)
    if not df:
        return
    df = pd.DataFrame(df)
    df = df.pivot(index='date', columns='horizon', values="bias").sort_index()
    return df


HORIZON_CONTAINER = ['daily', 'monthly', 'quarterly']


@export_as_api
def get_index_factor_exposure(
    order_book_ids, start_date=None, end_date=None, factors=None, market="cn"
):
    """获取因子暴露度

    :param order_book_ids: 股票代码或代码列表
    :param start_date: 如'2013-01-04' (Default value = None)
    :param end_date: 如'2014-01-04' (Default value = None)
    :param factors: 如'yield', 'beta', 'volatility' (Default value = None)
    :param market: 地区代码, 如'cn' (Default value = "cn")
    """
    try:
        order_book_ids = ensure_order_book_ids(order_book_ids, type="INDX")
    except ValueError:
        return

    start_date, end_date = ensure_date_range(start_date, end_date)

    if factors is not None:
        factors = ensure_list_of_string(factors)
        check_items_in_container(factors, exposure_factors, "factors")

    results = get_client().execute(
        "get_index_factor_exposure", order_book_ids, start_date, end_date, factors, market=market
    )

    if not results:
        return None
    df = pd.DataFrame.from_records(results, index=['date', 'order_book_id'])
    df.sort_index(inplace=True)
    return df



# -*- coding: utf-8 -*-
import socket
from collections import deque
from threading import Lock, Semaphore
from contextlib import contextmanager

from rqdatac.share.errors import ErrorFromBackend, GatewayError
from rqdatac.decorators import retry
from rqdatac.connection import Connection
from rqdatac.utils import connection_error, timeout_error


class ConnectionPool:
    def __init__(self, addr, auth=None, max_pool_size=8, connect_timeout=3.0, timeout=5 * 60):
        self._addr = addr
        self._auth = auth if auth is not None else {}
        self._connect_timeout = connect_timeout
        self._timeout = timeout
        self._lock = Lock()
        self._free_connections = deque()
        self._semaphore = Semaphore(max_pool_size)

    @retry(10, suppress_exceptions=connection_error)
    def execute(self, method, *args, **kwargs):
        with self._semaphore:
            with self._get_connection() as conn:
                return conn.execute(method, *args, **kwargs)

    def reset(self):
        with self._lock:
            for conn in self._free_connections:
                conn.close()
            self._free_connections.clear()

    @contextmanager
    def _get_connection(self):
        conn = self._ensure_connection()
        try:
            yield conn
        except (KeyboardInterrupt, Exception) as e:
            if not isinstance(e, (ErrorFromBackend, GatewayError)):
                conn.close()
            raise e
        else:
            with self._lock:
                self._free_connections.append(conn)

    def _ensure_connection(self):
        with self._lock:
            while self._free_connections:
                # round-robin to distribute load
                conn = self._free_connections.popleft()
                if conn.is_normal():
                    return conn
                else:
                    conn.close()

        return self._new_connection()

    @retry(3, suppress_exceptions=timeout_error)
    def _new_connection(self):
        s = Connection.sock_factory(self._addr, timeout=self._connect_timeout)
        s.settimeout(self._timeout)
        return Connection(s, self._auth)

    close = reset



# -*- coding: utf-8 -*-
import os
import re
import socket
import warnings
import logging
from functools import partial, wraps

from six.moves.urllib.parse import urlparse
from six import string_types
from sys import platform
import rqdatac

URI_PATTERN = r"^(?P<schema>tcp|rqdatac?)://(?P<username>\S+?):(?P<password>\S+)@(?P<hostname>\S+):(?P<port>\d+$)"
URI_PATTERN = re.compile(URI_PATTERN)

if getattr(os, "register_at_fork", None):
    def _close_after_fork():
        global _CLIENT
        if _CLIENT is not _DUMMY and _CLIENT.PID != os.getpid():
            reset()


    os.register_at_fork(after_in_child=_close_after_fork)


class DummyClient:
    PID = -1

    def execute(self, *args, **kwargs):
        raise RuntimeError("rqdatac is not initialized")

    def reset(self):
        pass

    def info(self):
        print('rqdatac is not initialized')

    def close(self):
        pass

    execute_with_timeout = execute


_DUMMY = DummyClient()

_CLIENT = _DUMMY

_PLUGINS_IMPORTED = False


def reset(flush_cache=False):
    """
    reset all connections. this function is helpful when you fork from an connected client.

    :params flush_cache: boolean, flush all ttl cache, default False
    """
    _CLIENT.reset()
    _CLIENT.PID = os.getpid()
    if flush_cache:
        from rqdatac.decorators import ttl_cache_clear
        ttl_cache_clear()


def initialized():
    """
    is rqdatac already initialized ?

    :return: Bool
    """
    global _CLIENT, _DUMMY
    return _CLIENT is not _DUMMY


def set_sock_opts(func):
    @wraps(func)
    def wrap(*args, **kwargs):
        s = func(*args, **kwargs)
        s.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        try:
            if platform.startswith('linux'):
                # (unit:s)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 60)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 5 * 60)
            elif platform.startswith('darwin'):  # macos
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 60)
            elif platform.startswith('win'):  # windows
                # (unit:ms) since windows2000
                s.ioctl(socket.SIO_KEEPALIVE_VALS, (1, 5 * 60 * 1000, 60 * 1000))
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 20)
        except:
            pass
        return s

    return wrap


def _parse_proxy_url(proxy_url):
    if not proxy_url:
        return

    proxy = urlparse(proxy_url)
    return proxy.scheme, proxy.hostname or proxy.path, proxy.port, proxy.username, proxy.password


def _set_sock_factory(proxy_info):
    from rqdatac.connection import Connection
    if proxy_info is None:
        conn = socket.create_connection
    else:
        try:
            import socks
        except ImportError:
            raise RuntimeError("PySocks is required when use proxy. You can install it using `pip install PySocks` or "
                               "`pip install rqdatac[proxy]`")

        if not isinstance(proxy_info, tuple) or len(proxy_info) != 5:
            raise ValueError("expected a tuple like (proxy_type, host, port, user, password)")

        proxy_type, host, port, user, password = proxy_info
        if proxy_type.upper() == "HTTP":
            proxy_type = socks.PROXY_TYPE_HTTP
        elif proxy_type.upper() == "SOCKS4":
            proxy_type = socks.PROXY_TYPE_SOCKS4
        elif proxy_type.upper() == "SOCKS5":
            proxy_type = socks.PROXY_TYPE_SOCKS5
        else:
            raise ValueError("proxy type {} not supported yet. http, socks4 and socks5 proxy are supported".format(proxy_type))
        conn = partial(socks.create_connection, proxy_type=proxy_type, proxy_addr=host, proxy_port=port,
                       proxy_username=user, proxy_password=password)

    Connection.set_sock_factory(set_sock_opts(conn))


def init(username=None, password=None, addr=("rqdatad-pro.ricequant.com", 16011), *_, **kwargs):
    """initialize rqdatac.

    rqdatac connection is thread safe but not fork safe. Every thread have their own connection by
    default. you can set param 'use_pool' to True to use a connection pool instead.

    NOTE: if you are using rqdatac with python < 3.7 in a multi-process program, remember to call
    reset in child process.

    Environment Variable:
    RQDATAC_CONF / RQDATAC2_CONF: When init called with no argument, the value in RQDATAC2_CONF is used as uri, then RQDATAC_CONF.
    RQDATAC_PROXY: proxy info, e.g. http://username:password@host:port

    :param username: string
    :param password: string
    :param addr: ('127.0.0.1', 80) or '127.0.0.1:80'

    :keyword uri: keyword only. a uri like rqdata://username:password@host:port or tcp://username:password@host:port
    :keyword connect_timeout: socket connect connect timeout, default is 5 sec.
    :keyword timeout: socket time out, default is 60 sec.
    :keyword lazy: True by default, means "do not connect to server immediately".
    :keyword use_pool: use connection pool. default is False
    :keyword max_pool_size: max pool size, default is 8
    :keyword proxy_info: a tuple like (proxy_type, host, port, user, password) if use proxy, default is None
    :keyword auto_load_plugins: boolean, enable or disable auto load plugin, default to True.
    """
    extra_args = {k: kwargs.pop(k) for k in ('timeout', 'connect_timeout') if k in kwargs}
    proxy_info = kwargs.pop('proxy_info', None) or _parse_proxy_url(os.environ.get('RQDATAC_PROXY'))
    _set_sock_factory(proxy_info)

    logging.getLogger("rqdata").disabled = not kwargs.pop('debug', False)

    uri = kwargs.pop("uri", None)
    if not (username or password or uri):
        uri = os.environ.get("RQDATAC2_CONF") or os.environ.get("RQDATAC_CONF")
    if username and password and addr:
        addr = parse_address(addr)
    elif uri:
        r = URI_PATTERN.match(uri)
        if r is not None:
            username = username or r.group('username')
            password = password or r.group('password')
            addr = parse_address((r.group("hostname"), r.group("port")))
        else:
            raise ValueError("uri invalid")
    else:
        err = ValueError("username/password/addr or uri expected")
        if not username:
            raise err
        r = URI_PATTERN.match(username)
        # pass uri as first argument.
        if not r:
            raise err
        username = r.group('username')
        password = r.group('password')
        addr = parse_address((r.group("hostname"), r.group("port")))
        if not (username and password):
            raise err

    global _CLIENT
    if _CLIENT is not _DUMMY:
        warnings.warn("rqdatac is already inited. Settings will be changed.", stacklevel=0)
        reset()

    global _PLUGINS_IMPORTED
    if not _PLUGINS_IMPORTED:
        if kwargs.get("auto_load_plugins", True):
            _auto_import_plugin()
            _PLUGINS_IMPORTED = True

    auth_info = {'username': username, 'password': password, 'ver': rqdatac.__version__}
    if kwargs.pop("use_pool", False):
        from .connection_pool import ConnectionPool
        max_pool_size = kwargs.pop("max_pool_size", 8)
        _CLIENT = ConnectionPool(addr, auth=auth_info, max_pool_size=max_pool_size, **extra_args)
    else:
        from .thread_local import ThreadLocalConnection
        _CLIENT = ThreadLocalConnection(addr, auth=auth_info, **extra_args)

    _CLIENT.PID = os.getpid()

    def show_info():
        print('rqdatac version:', rqdatac.__version__)
        print('server address: {}:{}'.format(addr[0], addr[1]))
        if username == 'license':
            print('You are using license:\r\n{}'.format(password))
        elif username.startswith('rqpro.'):
            print('You are using a RQPro account: {}'.format(username.split('.', 1)[1]))
        elif username == 'sid':
            print('You are using your RQPro account')
        else:
            print('You are using account: {}'.format(username))

    _CLIENT.info = show_info

    if username == "license":
        quota = get_client().execute("user.get_quota")
        remaining_days = quota["remaining_days"]
        is_trial = quota["license_type"] == "TRIAL"
        if is_trial or 0 <= remaining_days <= 14:
            warnings.warn("Your account will be expired after  {} days. "
                          "Please call us at 0755-22676337 to upgrade or purchase or "
                          "renew your contract.".format(remaining_days))
    elif not kwargs.get("lazy", True):
        get_client().execute("get_all_trading_dates")


def get_client():
    if _CLIENT.PID != os.getpid():
        reset()
    return _CLIENT


def parse_address(address):
    if isinstance(address, tuple):
        host, port = address
        return host, int(port)
    elif isinstance(address, string_types):
        if ":" not in address:
            return address
        host, port = address.rsplit(":", 1)
        return host, int(port)
    else:
        raise ValueError("address must be a str or tuple")


def _auto_import_plugin():
    from pkgutil import iter_modules
    from importlib import import_module
    plugin_module_names = []
    for m in iter_modules():
        _, name, is_pkg = m
        if not is_pkg:
            continue
        if name.startswith('rqdatac_'):
            plugin_module_names.append(name)

    plugin_module_names.sort()
    for name in plugin_module_names:
        logging.info('loading plugin {}'.format(name))
        import_module(name)



# -*- coding: utf-8 -*-
from collections import defaultdict
import pandas as pd
import numpy as np

from rqdatac.validators import (
    ensure_list_of_string,
    ensure_order_book_ids,
    check_items_in_container,
    ensure_date_range
)

from rqdatac.client import get_client
from rqdatac.decorators import export_as_api


CONSENSUS_INDICATOR_FIELDS = [
    'net_profit_t', 'net_profit_t1', 'net_profit_t2',
    'revenue_t', 'revenue_t1', 'revenue_t2',
    'net_asset_t', 'net_asset_t1', 'net_asset_t2',
    'cash_from_operating_activities_t',
    'cash_from_operating_activities_t1',
    'cash_from_operating_activities_t2'
]


DTYPES = {k: '<f8' for k in CONSENSUS_INDICATOR_FIELDS}
DTYPES['fiscal_year'] = '<u8'

CONSENSUS_PRICE_FIELDS = [
    'half_year_target_price',
    'one_year_target_price',
    'quarter_recommendation',
    'half_year_recommendation',
    'one_year_recommendation',
]

CONSENSUS_PRICE_FIELDS_MAP = {
    'half_year_target_price': ('price_raw', 'price_prd', 'M06'),
    'one_year_target_price': ('price_raw', 'price_prd', 'Y01'),
    'quarter_recommendation': ('grd_coef', 'grd_prd', '1'),
    'half_year_recommendation': ('grd_coef', 'grd_prd', '2'),
    'one_year_recommendation': ('grd_coef', 'grd_prd', '3'),
}

PRICE_DTYPES = {
    'half_year_target_price': '<f8',
    'one_year_target_price': '<f8',
    'quarter_recommendation': '<f8',
    'half_year_recommendation': '<f8',
    'one_year_recommendation': '<f8',
}


@export_as_api
def get_consensus_indicator(order_book_ids, fiscal_year, fields=None, market='cn'):
    """
    获取一致预期数据

    :param order_book_ids: 股票名称
    :param fiscal_year: int/str, 查询年份
    :param fields: list,  一致预期字段
    :param market: (Default value = 'cn')
    :returns: pandas  MultiIndex DataFrame

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    fiscal_year = int(fiscal_year)
    if fields is None:
        fields = CONSENSUS_INDICATOR_FIELDS
    else:
        fields = ensure_list_of_string(fields, 'consensus_indicator')
        check_items_in_container(fields, CONSENSUS_INDICATOR_FIELDS, 'consensus_indicator')
    data = get_client().execute('get_consensus_indicator', order_book_ids, fiscal_year, fields, market='cn')
    if not data:
        return None
    df = pd.DataFrame(data)
    df.set_index(['order_book_id', 'date'], inplace=True)
    df.sort_index(inplace=True)
    dtypes = {f: DTYPES[f] for f in df.columns if f in DTYPES}
    df = df.astype(dtypes)
    return df


@export_as_api
def get_consensus_price(order_book_ids, start_date=None, end_date=None, fields=None, adjust_type='none', market='cn'):
    """
    获取一致预期股价预测数据

    :param order_book_ids: 股票名称
    :param start_date: 开始日期， date-like object, 默认三月前那天
    :param end_date: 结束日期， date-like object， 默认当天
    :param fields: list,  一致预期字段
    :param adjust_type: 可选参数,默认为‘none', 返回原始数据
            'pre' 返回前复权数据
            'post' 返回后复权数据
    :param market: (Default value = 'cn')
    :returns: pandas MultiIndex DataFrame

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)

    if fields is None:
        fields = CONSENSUS_PRICE_FIELDS
    else:
        fields = ensure_list_of_string(fields, 'consensus_price')
        check_items_in_container(fields, CONSENSUS_PRICE_FIELDS, 'consensus_price')

    start_date, end_date = ensure_date_range(start_date, end_date)

    data = get_client().execute('get_consensus_price', order_book_ids, start_date, end_date, market='cn')
    if not data:
        return None

    records = defaultdict(dict)
    for r in data:
        key = (r['order_book_id'], r['institute'], r['date'])
        if key not in records:
            records[key].update(r)
        for field in fields:
            name, f, value = CONSENSUS_PRICE_FIELDS_MAP[field]
            if r[f] == value:
                records[key][field] = r[name]
    df = pd.DataFrame(list(records.values()))
    df.set_index(['order_book_id', 'date'], inplace=True)
    df.sort_index(inplace=True)
    df = df.reindex(columns=['institute'] + fields)
    df = df.astype({f: PRICE_DTYPES[f] for f in fields})

    if adjust_type != 'none':
        adjust_fields = list(set(CONSENSUS_PRICE_FIELDS[:2]) & set(fields))
        if not adjust_fields:
            return df

        from rqdatac.services.detail.adjust_price import get_ex_factor_for
        ex_factors = get_ex_factor_for(order_book_ids, market)
        pre = adjust_type == 'pre'

        def adjust(sub):
            factors = ex_factors.get(sub.name)
            if factors is None:
                return sub
            factor = np.take(factors.values, factors.index.searchsorted(sub.index.get_level_values(1), side='right') - 1)
            if pre:
                factor /= factors.iloc[-1]

            sub[adjust_fields] = (sub[adjust_fields].values.T * factor).T
            return sub

        df = df.groupby(level=0).apply(adjust)
        df[adjust_fields] = np.round(df[adjust_fields], 4)

    return df



# -*- coding: utf-8 -*-
import warnings
from collections import OrderedDict

import pandas as pd

from rqdatac.validators import (
    ensure_list_of_string,
    ensure_order,
    check_items_in_container,
    ensure_date_range,
    ensure_date_int,
    ensure_order_book_ids,
    raise_for_no_panel,
)
from rqdatac.client import get_client
from rqdatac.decorators import export_as_api, compatible_with_parm
from rqdatac.utils import pf_fill_nan, is_panel_removed


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_split(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取拆分信息

    :param order_book_ids: 股票 order_book_id or order_book_id list
    :param start_date: 开始日期；默认为上市首日
    :param end_date: 结束日期；默认为今天
    :param market:  (Default value = "cn")

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    if start_date is not None:
        start_date = ensure_date_int(start_date)
    if end_date is not None:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("get_split", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    df.sort_values("ex_dividend_date", inplace=True)
    # cumprod [1, 2, 4] -> [1, 1*2, 1*2*4]
    df["cum_factor"] = df["split_coefficient_to"] / df["split_coefficient_from"]
    df["cum_factor"] = df.groupby("order_book_id")["cum_factor"].cumprod()
    if len(order_book_ids) == 1:
        df.set_index("ex_dividend_date", inplace=True)
    else:
        df.set_index(["order_book_id", "ex_dividend_date"], inplace=True)
    df.sort_index(inplace=True)
    return df


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_dividend(order_book_ids, start_date=None, end_date=None, adjusted=False, market="cn"):
    """获取分红信息

    :param order_book_ids: 股票 order_book_id or order_book_id list
    :param start_date: 开始日期，默认为股票上市日期
    :param end_date: 结束日期，默认为今天
    :param adjusted: deprecated
    :param market:  (Default value = "cn")

    """
    if adjusted:
        warnings.warn(
            "get_dividend adjusted = `True` is not supported yet. "
            "The default value is `False` now."
        )
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    if start_date is not None:
        start_date = ensure_date_int(start_date)
    if end_date is not None:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("get_dividend", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    if len(order_book_ids) == 1:
        df.set_index("declaration_announcement_date", inplace=True)
    else:
        df.set_index(["order_book_id", "declaration_announcement_date"], inplace=True)
    return df.sort_index()


@export_as_api
def get_dividend_info(order_book_ids, start_date=None, end_date=None, market="cn"):
    """对应时间段是否发生分红

    :param order_book_ids: 股票 order_book_id or order_book_id list
    :param start_date: 开始日期，默认为空
    :param end_date: 结束日期，默认为空
    :param market:  (Default value = "cn")

    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    if start_date is not None:
        start_date = ensure_date_int(start_date)
    if end_date is not None:
        end_date = ensure_date_int(end_date)
    if start_date and end_date:
        if start_date > end_date:
            raise ValueError("invalid date range: [{!r}, {!r}]".format(start_date, end_date))

    data = get_client().execute("get_dividend_info", order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)
    if len(order_book_ids) == 1:
        df.set_index("effective_date", inplace=True)
    else:
        df.set_index(["order_book_id", "effective_date"], inplace=True)
    return df.sort_index()


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_ex_factor(order_book_ids, start_date=None, end_date=None, market="cn"):
    """获取复权因子

    :param order_book_ids: 如'000001.XSHE'
    :param market: 国家代码, 如 'cn' (Default value = "cn")
    :param start_date: 开始日期，默认为股票上市日期
    :param end_date: 结束日期，默认为今天
    :returns: 如果有数据，返回一个DataFrame, 否则返回None

    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)
    if start_date is not None:
        start_date = ensure_date_int(start_date)
    if end_date is not None:
        end_date = ensure_date_int(end_date)
    data = get_client().execute("get_ex_factor", order_book_ids, start_date, end_date, market=market)
    if not data:
        return None
    df = pd.DataFrame(data)
    df.sort_values(["order_book_id", "ex_date"], inplace=True)
    df.set_index("ex_date", inplace=True)
    return df


TURNOVER_FIELDS_MAP = OrderedDict()
TURNOVER_FIELDS_MAP["today"] = "turnover_rate"
TURNOVER_FIELDS_MAP["week"] = "week_turnover_rate"
TURNOVER_FIELDS_MAP["month"] = "month_turnover_rate"
TURNOVER_FIELDS_MAP["year"] = "year_turnover_rate"
TURNOVER_FIELDS_MAP["current_year"] = "year_sofar_turnover_rate"


def _get_maped_fields(fields):
    fields = ensure_list_of_string(fields, "fields")
    check_items_in_container(fields, TURNOVER_FIELDS_MAP, "fields")
    fields = ensure_order(fields, TURNOVER_FIELDS_MAP.keys())
    return fields, [TURNOVER_FIELDS_MAP[field] for field in fields]


@export_as_api
def get_turnover_rate(order_book_ids, start_date=None, end_date=None, fields=None, expect_df=True, market="cn"):
    """获取股票换手率数据

    :param order_book_ids: 股票代码或股票代码列表
    :param start_date: 开始时间
    :param end_date: 结束时间；在 start_date 和 end_date 都不指定的情况下，默认为最近3个月
    :param fields: str或list类型. 默认为None, 返回所有fields.
                   field 包括： 'today', 'week', 'month', 'year', 'current_year'
                   (Default value = None)
    :param expect_df: 返回 MultiIndex DataFrame (Default value = True)
    :param market: 地区代码, 如: 'cn' (Default value = "cn")
    :returns: 如果order_book_ids或fields为单个值 返回pandas.DataFrame, 否则返回pandas.Panel

    """
    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)
    if fields is not None:
        fields, mapped_fields = _get_maped_fields(fields)
    else:
        fields, mapped_fields = list(TURNOVER_FIELDS_MAP.keys()),  list(TURNOVER_FIELDS_MAP.values())
    df = get_client().execute(
        "get_turnover_rate", order_book_ids, start_date, end_date, mapped_fields, market=market
    )
    if not df:
        return
    df = pd.DataFrame(df, columns=["tradedate", "order_book_id"] + mapped_fields)
    df.rename(columns={v: k for k, v in TURNOVER_FIELDS_MAP.items()}, inplace=True)

    if not expect_df and not is_panel_removed:
        df.set_index(["tradedate", "order_book_id"], inplace=True)
        df.sort_index(inplace=True)
        df = df.to_panel()
        df = pf_fill_nan(df, order_book_ids)
        if len(order_book_ids) == 1:
            df = df.minor_xs(*order_book_ids)
            if fields and len(fields) == 1:
                return df[fields[0]]
            return df
        if fields and len(fields) == 1:
            return df[fields[0]]
        warnings.warn("Panel is removed after pandas version 0.25.0."
                      " the default value of 'expect_df' will change to True in the future.")
        return df
    else:
        df.sort_values(["order_book_id", "tradedate"], inplace=True)
        df.set_index(["order_book_id", "tradedate"], inplace=True)
        if expect_df:
            return df

        if len(order_book_ids) != 1 and len(fields) != 1:
            raise_for_no_panel()

        if len(order_book_ids) == 1:
            df.reset_index(level=0, drop=True, inplace=True)
            if len(fields) == 1:
                df = df[fields[0]]
            return df
        else:
            df = df.unstack(0)[fields[0]]
            df.index.name = None
            df.columns.name = None
            return df


@export_as_api
def get_price_change_rate(order_book_ids, start_date=None, end_date=None, expect_df=True, market="cn"):
    """获取价格变化信息

    :param order_book_ids: 股票列表
    :param start_date: 开始日期: 如'2013-01-04'
    :param end_date: 结束日期: 如'2014-01-04'；在 start_date 和 end_date 都不指定的情况下，默认为最近3个月
    :param expect_df: 是否返回 MultiIndex DataFrame (Default value = True)
    :param market: 地区代码
    :returns: 如果输入一只股票, 则返回pandas.Series, 否则返回pandas.DataFrame

    """
    start_date, end_date = ensure_date_range(start_date, end_date)
    order_book_ids = ensure_order_book_ids(order_book_ids)
    data = get_client().execute("get_daily_returns", order_book_ids, start_date, end_date, market=market)
    if not data:
        return None
    df = pd.DataFrame(data)

    if len(order_book_ids) == 1 and not expect_df:
        df = df.set_index("date")
        series = df["daily_return"]
        series.name = order_book_ids[0]
        return series
    df = df.pivot(index="date", columns="order_book_id", values="daily_return")
    return df.sort_index()


@export_as_api
@compatible_with_parm(name="country", value="cn", replace="market")
def get_yield_curve(start_date=None, end_date=None, tenor=None, market="cn"):
    """获取国债收益率曲线

    :param market: 地区代码, 如'cn', 'us' (Default value = "cn")
    :param start_date: 开始日期 (Default value = "2013-01-04")
    :param end_date: 结束日期 (Default value = "2014-01-04")
    :param tenor: 类别, 如 OS, 1M, 3M, 1Y (Default value = None)

    """
    start_date, end_date = ensure_date_range(start_date, end_date)
    all_tenor = (
        "0S",
        "1M",
        "2M",
        "3M",
        "6M",
        "9M",
        "1Y",
        "2Y",
        "3Y",
        "4Y",
        "5Y",
        "6Y",
        "7Y",
        "8Y",
        "9Y",
        "10Y",
        "15Y",
        "20Y",
        "30Y",
        "40Y",
        "50Y",
    )
    if tenor:
        tenor = ensure_list_of_string(tenor, "tenor")
        check_items_in_container(tenor, all_tenor, "tenor")
        tenor = ensure_order(tenor, all_tenor)
    df = get_client().execute("get_yield_curve", start_date, end_date, tenor, market=market)
    if not df:
        return
    columns = ["trading_date"]
    columns.extend(tenor or all_tenor)
    df = pd.DataFrame(df, columns=columns)
    df.set_index("trading_date", inplace=True)
    return df.sort_index()


@export_as_api
def get_block_trade(order_book_ids, start_date=None, end_date=None, market='cn'):
    """获取大宗交易信息
    :param order_book_ids: 股票代码
    :param start_date: 起始日期，默认为前三个月
    :param end_date: 截止日期，默认为今天
    :param market: (default value = 'cn')
    :return: pd.DataFrame or None
    """

    order_book_ids = ensure_order_book_ids(order_book_ids)
    start_date, end_date = ensure_date_range(start_date, end_date)

    data = get_client().execute('get_block_trade', order_book_ids, start_date, end_date, market=market)
    if not data:
        return
    df = pd.DataFrame(data)[['order_book_id', 'trade_date', 'price', 'volume', 'total_turnover', 'buyer', 'seller']]
    df.set_index(["order_book_id", "trade_date"], inplace=True)
    df.sort_index(inplace=True)
    return df

EXCHANGE_DATE_FIELDS=[
    "currency_pair",
    "bid_referrence_rate",
    "ask_referrence_rate",
    "middle_referrence_rate",
    "bid_settlement_rate_sh",
    "ask_settlement_rate_sh",
    "bid_settlement_rate_sz",
    "ask_settlement_rate_sz",
]

@export_as_api
def get_exchange_rate(start_date=None, end_date=None, fields=None):
    """获取汇率信息

    :param start_date: 开始日期, 如 '2013-01-04' (Default value = None)
    :param end_date: 结束日期, 如 '2014-01-04' (Default value = None)
    :param fields: str or list 返回 字段名称:currency_pair、bid_referrence_rate、ask_referrence_rate、middle_referrence_rate
        bid_settlement_rate_sh、ask_settlement_rate_sh、bid_settlement_rate_sz、ask_settlement_rate_sz

    """
    start_date, end_date = ensure_date_range(start_date, end_date)
    if fields:
        fields = ensure_list_of_string(fields, "fields")
        check_items_in_container(fields, EXCHANGE_DATE_FIELDS, "fields")
    else:
        fields = EXCHANGE_DATE_FIELDS

    data = get_client().execute("get_exchange_rate", start_date, end_date, fields)
    if not data:
        return None
    df = pd.DataFrame(data)
    df.set_index("date", inplace=True)
    df = df[fields]
    return df


TEMPORARY_CODE_FIELDS = [
    "symbol",
    "temporary_trade_code",
    "temporary_symbol",
    "temporary_round_lot",
    "temporary_effective_date",
    "parallel_effective_date",
    "parallel_cancel_date"
]


@export_as_api
def get_temporary_code(order_book_ids, market="cn"):
    """临时交易代码查询

    :param order_book_ids: 股票 order_book_id or order_book_id list
    :param market:  (Default value = "cn")
    """
    order_book_ids = ensure_order_book_ids(order_book_ids, market=market)

    data = get_client().execute("get_temporary_code", order_book_ids, market)
    if not data:
        return None
    df = pd.DataFrame(data)
    df.set_index("order_book_id", inplace=True)
    df = df[TEMPORARY_CODE_FIELDS]
    return df